/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  deadlock.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/10/25 13:46:01 $
 * $Log: deadlock.c,v $
 * Revision 1.2  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend checks an execution of a parallel program
 * for the occurence of deadlocks and livelocks and warns you
 * *before* the lock occurs.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// DEADLOCK ET AL DETECTIE

// MR. 9/00


// type 1: a wacht op b, b wacht op c, c wacht op ... , ... wacht op a
//    werkt ook met speciaal geval a wacht op a
// type 2: a heeft lock L1, a stopt ermee (al dan niet met pthread_exit)
//      2a: andere draad wil L1 nu
//      2b: andere draad wil L1 in de toekomst
//   => notie van dode draad invoeren + warning op scherm
// probleem 3: draad b geeft lock van draad a vrij (met a eventueel DEAD_THREAD)
// probleem 4: draad b geeft vrije lock vrij
// probleem 5: thr_join met geblokkeerde draad

// Waarschuwt ook voor:
//   -draad stopt zonder pthread_exit() uit te voeren

#include "jiti.h"
#include <errno.h>   // voor EBUSY

/******************

Mogelijke status voor een draad:

-onbestaand
-CREATED
-RUNNING
-EXITED
-DEADLOCKING: net voor DEADLOCKED
-DEADLOCKED
-JOINING

per draad:
-if JOINING -> met wie
-lijst van LOCKS in bezit

*****************/


#define NO_THREAD   -1
#define DEAD_THREAD -2


#define AANTAL_INSTR_FUNC     5
#define PTHREAD_MUTEX_LOCK    0
#define PTHREAD_MUTEX_UNLOCK  1
#define PTHREAD_EXIT          2
#define PTHREAD_CREATE        3
#define PTHREAD_JOIN          4

int  dl_pthread_mutex_lock(pthread_mutex_t *mutex);
int  dl_pthread_mutex_unlock(pthread_mutex_t *mutex);
void dl_pthread_exit(void * retval);
int  dl_pthread_create(pthread_t * thread, const pthread_attr_t * attr, 
		       void * (*start_routine)(void *), void* arg);  
int  dl_pthread_join(pthread_t th, void **thread_return); 

struct _reloc_info_ {
  char function[50];
  int(*real_address)(pthread_mutex_t*);
  unsigned instr_address;
} reloc_info[AANTAL_INSTR_FUNC] = {
  {"pthread_mutex_lock",      0, (unsigned)dl_pthread_mutex_lock},     // 0
  {"pthread_mutex_unlock",    0, (unsigned)dl_pthread_mutex_unlock},   // 1
  {"pthread_exit",            0, (unsigned)dl_pthread_exit},           // 2
  {"pthread_create",          0, (unsigned)dl_pthread_create},         // 3
  {"pthread_join",            0, (unsigned)dl_pthread_join}            // 4
};

static nr_threads_running=1;
static nr_threads_created=1;
static nr_threads_blocked=0;


#define MAX_NR_THREADS 1024
static int thread_info[MAX_NR_THREADS];
#define T_NEVER_EXISTED 0   // default
#define T_RUNNING       1
#define T_DEADLOCKED    2
#define T_EXITED        3
#define T_JUST_DIED     4   // deadlocked, maar moet nog afgehandeld worden
#define T_JOIN(X)      -X   // thread_info[] == -t => aan het joinen met t

///////////////////////////////////////////

typedef struct _thread_locks_info_ {
  char * lock_address;
  int threadid;
  int last_threadid;
  unsigned ip;
  struct _thread_locks_info_ * next;
} thread_locks_info_t;

static thread_locks_info_t * hold_locks_info = NULL;
static thread_locks_info_t * wanted_locks_info = NULL;

///////INTERN///////////////////////////////

static int _MR_INTERN_mutex_lock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_LOCK].real_address ?
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex) :
    pthread_mutex_lock(mutex) ;
}

static int _MR_INTERN_mutex_unlock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_UNLOCK].real_address ?
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex) :
    pthread_mutex_unlock(mutex) ;
}                          
                 
static pthread_mutex_t _intern_mutex_;

#define LOCK()     _MR_INTERN_mutex_lock(&_intern_mutex_)
#define UNLOCK()   _MR_INTERN_mutex_unlock(&_intern_mutex_)
#define ATOMIC(X)  LOCK();X;UNLOCK()

//////////////////////////////////////////


void give_info(int deadlock){
  //  LOCK();
  printf("Threads created: %d, running: %d, blocked forever %d", 
	 nr_threads_created, nr_threads_running, nr_threads_blocked);
  if (deadlock) printf(", %d deadlocked", deadlock);
  printf("\n");  
  if (!nr_threads_running) printf("---\n\nThere are no running threads left!!!\n");
  //  UNLOCK();
}


/////////////////////////////////////////


PRINT(thread_locks_info_t * lock_info){
  while(lock_info){
    fprintf(jiti_tty, "[0x%x T%d]\n", lock_info->lock_address, lock_info->threadid);
    lock_info=lock_info->next;
  }
}


/////////////////////////////////////////


unsigned ip2source(unsigned ip){
  if (ip>0x40000000) return 0; // 't is gedaan, we zijn in jiti gearriveerd
  jiti_addr2line(ip);
  return 1;  // op naar het volgende frame
}    


/////////////////////////////////////////


static owner(int threadid, char * object, unsigned ip){
  thread_locks_info_t * thread_locks_info = malloc(sizeof(thread_locks_info_t));
  thread_locks_info->lock_address = object;
  thread_locks_info->threadid     = threadid;
  thread_locks_info->ip           = ip-delta;
  LOCK();
  thread_locks_info->next         = hold_locks_info;
  hold_locks_info                 = thread_locks_info;
  fprintf(jiti_tty, "---> T%d now owns 0x%x\nHELD:\n", threadid, object);
  PRINT(hold_locks_info);
  UNLOCK();
}


//////////////////////////////////////////


static thread_locks_info_t * who_owns(char * object){
  thread_locks_info_t * thread_locks_info = hold_locks_info;
  while (thread_locks_info){
    if (thread_locks_info->lock_address==object) return thread_locks_info;
    thread_locks_info = thread_locks_info->next;
  }
  return NULL;  // not found => nobody holds the lock
}


//////////////////////////////////////////


static noowner(int threadid, char * object){
  thread_locks_info_t * thread_locks_info      = hold_locks_info;
  thread_locks_info_t * thread_locks_info_prev = NULL;
  LOCK();
  if (!thread_locks_info){
    printf("---\nThread %d is trying to unlock the free object 0x%x\n", threadid, object);
    print_my_stack(ip2source, 2);
    PRINT(hold_locks_info);
  } else if (!thread_locks_info->next) {
    // maar 1 element in lijst
    if (thread_locks_info->lock_address!=object){
      fprintf(jiti_tty, "PANIC: wrong hold_locks_info!\n");
      PRINT(hold_locks_info);
    } else {  // verwijder enige element
      if (thread_locks_info->threadid!=threadid) {
	printf("---\nThread %d is unlocking object %x that is owned by %sthread %d\n",
	       threadid, object, 
	       thread_locks_info->threadid==DEAD_THREAD ? "dead " : "",
	       thread_locks_info->threadid==DEAD_THREAD ? thread_locks_info->last_threadid : 
	       thread_locks_info->threadid);
	print_my_stack(ip2source, 2);	
      } 
      free(thread_locks_info);
      hold_locks_info=NULL;      
    }      
  } else {  // minstens 2 elementen 
    if (thread_locks_info->lock_address==object){
      if (thread_locks_info->threadid!=threadid) {
	printf("---\nThread %d is unlocking object %x that is owned by thread %d\n",
	       threadid, object, thread_locks_info->threadid);
      }       
      // eerste element moet weg
      hold_locks_info = thread_locks_info->next;
      free(thread_locks_info);      
    } else {
      // nu pas kunnen we met 2 pointers lopen
      while (thread_locks_info){
	if (thread_locks_info->lock_address==object){
	  if (thread_locks_info->threadid!=threadid) {
	    printf("---\nThread %d is unlocking object %x that is owned by thread %d\n",
		   threadid, object, thread_locks_info->threadid);
	    print_my_stack(ip2source, 2);
	  } 
	  thread_locks_info_prev->next=thread_locks_info->next;
	  free(thread_locks_info);
	}
	thread_locks_info_prev = thread_locks_info;
	thread_locks_info = thread_locks_info->next;
      }
    }
  }
  fprintf(jiti_tty, "---> T%d no longer owns 0x%x\nHELD:\n", threadid, object);
  PRINT(hold_locks_info);
  UNLOCK();
}


//////////////////////////////////////////


static wants(int threadid, char * object, unsigned ip){
  thread_locks_info_t * thread_locks_info = malloc(sizeof(thread_locks_info_t));
  thread_locks_info->lock_address = object;
  thread_locks_info->threadid     = threadid;
  thread_locks_info->ip           = ip-delta;
  //  LOCK(); gebruikt nu lock die door check_deadlock() genomen wordt.
  thread_locks_info->next         = wanted_locks_info;
  wanted_locks_info               = thread_locks_info;
  fprintf(jiti_tty, "---> T%d wants 0x%x\nWANTED:\n", threadid, object);
  PRINT(wanted_locks_info);
  //  UNLOCK();
}


//////////////////////////////////////////


static char * what_wants(int threadid){
  thread_locks_info_t * thread_locks_info = wanted_locks_info;

  if ((threadid==NO_THREAD) || (threadid==DEAD_THREAD)) 
    return NULL;

  while (thread_locks_info){
    if (thread_locks_info->threadid==threadid) return thread_locks_info->lock_address;
    thread_locks_info = thread_locks_info->next;
  }
  return NULL;  // not found => nobody holds the lock
}


/////////////////////////////////////////


static dontwant(int threadid, char * object){
  thread_locks_info_t * thread_locks_info      = wanted_locks_info;
  thread_locks_info_t * thread_locks_info_prev = NULL;
  LOCK();
  if (!thread_locks_info){
    fprintf(jiti_tty, "PANIC: NO thread_locks_info!\n");
    PRINT(wanted_locks_info);
  } else if (!thread_locks_info->next) {
    // maar 1 element in lijst
    if ((thread_locks_info->threadid!=threadid) || (thread_locks_info->lock_address!=object)){
      fprintf(jiti_tty, "PANIC: WRONG thread_locks_info!\n");
      PRINT(wanted_locks_info);
    } else {  // verwijder enige element
      free(thread_locks_info);
      wanted_locks_info=NULL;      
    }      
  } else {  // minstens 2 elementen 
    if ((thread_locks_info->threadid==threadid) || (thread_locks_info->lock_address==object)){
      // eerste element moet weg
      wanted_locks_info = thread_locks_info->next;
      free(thread_locks_info);      
    } else {
      // nu pas kunnen we loopen met 2 pointers
      while (thread_locks_info){
	if ((thread_locks_info->threadid==threadid) || (thread_locks_info->lock_address==object)){
	  thread_locks_info_prev->next=thread_locks_info->next;
	  free(thread_locks_info);
	}
	thread_locks_info_prev = thread_locks_info;
	thread_locks_info = thread_locks_info->next;
      }
    }
  }
  fprintf(jiti_tty, "---> T%d doesn't want 0x%x any longer\nWANTED:\n", threadid, object);
  PRINT(wanted_locks_info);
  UNLOCK();
}


//////////////////////////////////////////


static dying(int threadid, int lock){
  thread_locks_info_t * thread_locks_info = hold_locks_info;  
  
  fprintf(jiti_tty, "Thread %d is dying.\n", threadid);

  if (lock) LOCK();
  // Zoek de locks die nog in het bezit zijn van de draad

  fprintf(jiti_tty, "HELD:\n");
  PRINT(hold_locks_info);

  while (thread_locks_info) {
    if (thread_locks_info->threadid==threadid){
      
      printf("---\nThread %d is %s and still owns the following lock: 0x%x\n", threadid,
      	     lock ? "exiting" : "blocked forever", thread_locks_info->lock_address);
      ip2source(thread_locks_info->ip);
      
      // noteer dat het object gelocked is door niemand  en nooit meer vrijkomt
      thread_locks_info->last_threadid = thread_locks_info->threadid;
      thread_locks_info->threadid = DEAD_THREAD;
      
      // wacht er iemand op het object?
      {
	thread_locks_info_t * thread_locks_info2 = wanted_locks_info;
	while (thread_locks_info2){
	  if (thread_locks_info2->lock_address==thread_locks_info->lock_address){
	    printf("Thread %d is waiting for this lock and is now considered dead.\n", 
		   thread_locks_info2->threadid);	    
	    // op zijn beurt is deze draad nu 'dood'.
	    nr_threads_blocked++;
	    give_info(0);
	    thread_info[thread_locks_info2->threadid] = T_DEADLOCKED;
	    dying(thread_locks_info2->threadid, 0);
	  }
	  thread_locks_info2 = thread_locks_info2->next;
	}
      }

    }
    thread_locks_info=thread_locks_info->next;
  }

  if (lock) UNLOCK();
}


//////////////////////////////////////////



static check_deadlock(int threadid, char * object, int info, unsigned ip){
  int ori_threadid=threadid;
  char * ori_object=object;
  int nr_threads_deadlocked=0;
  thread_locks_info_t * owner;

  
  if (!info) LOCK();  // we hebben de lock nog (en geven ze nooit meer weg...)

  if (info) printf("---\nDeadlock detected!\n");

  if (info) {
    printf("Thread %d wants to lock 0x%x:\n", threadid, object);
    thread_info[threadid]=T_JUST_DIED;
    print_my_stack(ip2source, 3);
  } else
    fprintf(jiti_tty, "---> Check for deadlock: T%d wants 0x%x\n", threadid, object);
  
  while (1){

    if (info) nr_threads_deadlocked++;

    owner = who_owns(object);

    if (!owner) goto OK; // niemand heeft de lock, enkel mogelijk indien hij net werd vrijgegeven
    
    threadid = owner->threadid;

    if (info){
      printf("This lock is held by thread %d:\n", threadid);
      ip2source(owner->ip);
    } else
      fprintf(jiti_tty, "---> The lock is held by T%d\n", threadid);
    

    if (threadid==DEAD_THREAD){
      printf("---\nThread %d wants to lock object 0x%x:\n", ori_threadid, ori_object);
      print_my_stack(ip2source, 2);
      printf("However, this object is locked by dead thread %d:\n", owner->last_threadid);
      ip2source(owner->ip);
      printf("As the object will never be released this thread is now considered dead.\n");
      nr_threads_blocked++;
      nr_threads_running--;
      give_info(0);
      dying(ori_threadid, 0);
      thread_info[ori_threadid]=T_JUST_DIED;      
      goto OK;
    }

    if (threadid==ori_threadid) {  // einde v/d lus bereikt
      // DEADLOCK

      if (!info) {
	// Opnieuw, maar schrijf nu info uit
	check_deadlock(ori_threadid, ori_object, 1, ip);
      } else {
	//	nr_threads_running -= nr_threads_deadlocked;
	printf("Threads created: %d, running: %d, blocked forever %d, deadlocked: %d\n", 
	       nr_threads_created, nr_threads_running, nr_threads_blocked, nr_threads_deadlocked);
	//	nr_threads_blocked += nr_threads_deadlocked;
	//	if (!nr_threads_running)
	//	  printf("There are no running threads left!!\n");
	UNLOCK();
	// Nu lopen we in de deadlock... Goodbye!
      }
    }
    
    object = what_wants(threadid);
    if (!object){
      fprintf(jiti_tty, "---> T%d isn't waiting for anything => OK\n", threadid);
      goto OK;
    } else {
      if (info){
	printf("Thread %d wants to lock 0x%x, ", threadid, object);
	thread_info[threadid]=T_JUST_DIED;
      } else
	fprintf(jiti_tty, "---> T%d waiting for 0x%x => DEADLOCK???\n", threadid, object);       
    }
    
  }
  
 OK:

  if (!info) wants(ori_threadid, ori_object, ip);

  // overloop de T_JUST_DIED-draden  
  {
    int t1, t2;
    for (t1=0; t1<MAX_NR_THREADS; t1++){
      if (thread_info[t1]==T_JUST_DIED){
	thread_info[t1] = T_DEADLOCKED;
	nr_threads_running--;
	nr_threads_blocked++;
	printf("---\nThread %d is deadlocked and is now considered dead.\n", t1);
	give_info(0);
	for (t2=0; t2<MAX_NR_THREADS; t2++){
	  if (thread_info[t2] == T_JOIN(t1)){
	    printf("---\nThread %d wanted to join with dead thread %d and is now considered dead.\n",
		   t2, t1);
	    thread_info[t2] = T_DEADLOCKED;
	    nr_threads_blocked++;
	    nr_threads_running--;
	    give_info(0);
	    dying(t2, 0);
	  }
	}
      }
    }
  }

  UNLOCK();
}


//////////////////////////////////////////


int dl_pthread_mutex_lock(pthread_mutex_t *mutex){
  int status;

  fprintf(jiti_tty, "---> LOCK %x by %d\n", mutex, THREADID);

  // we proberen eerst de lock te verkrijgen m.b.v. trylock
  status = pthread_mutex_trylock(mutex);

  // Mogelijkheid 1: we hebben de lock => geen deadlock, enkel noteren dat 'THREADID' nu
  //                 eigenaar is van 'mutex'
  if (status!=EBUSY) {
    owner(THREADID, (char*)mutex, get_ip());
    return status;
  }

  // Mogelijkheid 2: we hebben de lock niet => test op deadlock
  check_deadlock(THREADID, (char*)mutex, 0, get_ip());

  // komen we hier => geen deadlock, we wachten dus gewoon op de mutex
  //  wants(THREADID, (char*)mutex); Gebeurt nu in check_deadlock()
  status = (*reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex);
  dontwant(THREADID, (char*)mutex);
  owner(THREADID, (char*)mutex, get_ip());
  return status;
}               


//////////////////////////////////////////


int dl_pthread_mutex_unlock(pthread_mutex_t *mutex){

  fprintf(jiti_tty, "---> UNLOCK %x by %d\n", mutex, THREADID);

  noowner(THREADID, (char*)mutex);

  /* UNLOCK */
  return (*reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex);
}               


//////////////////////////////////////////


void dl_pthread_exit(void * retval){
  fprintf(jiti_tty, "Thread %d exiting.\n", THREADID);

  dying(THREADID, 1);

  ATOMIC(nr_threads_running--);

  thread_info[THREADID] = T_EXITED;

  (*reloc_info[PTHREAD_EXIT].real_address)(retval);

  // never reached
}


//////////////////////////////////////////


int dl_pthread_join(pthread_t th, void **thread_return){
  int status;
  int join_th=TID(th);
  fprintf(jiti_tty, "Thread %d wants to join with %d.\n", THREADID, join_th);

  if (thread_info[join_th]==T_DEADLOCKED){
    thread_info[THREADID]=T_DEADLOCKED;
    dying(THREADID, 1);
    ATOMIC(nr_threads_blocked++);
    ATOMIC(nr_threads_running--);
    printf("---\nThread %d wants to join with deadlocked thread %d.\n"
	   "This threads is now considered dead.\n", THREADID, join_th);
    give_info(0);
  }

  thread_info[THREADID] = T_JOIN(join_th);

  status =  (*(int(*)(pthread_t, void**))reloc_info[PTHREAD_JOIN].real_address)(th, thread_return);
  
  return status;
}


//////////////////////////////////////////


struct thread_info_t {
  void * (*start_routine)(void *);
  void * arg;
};    


//////////////////////////////////////////


void dl_pthread_create_helper(struct thread_info_t * helper_thread_info){
  fprintf(jiti_tty, "Thread %d running.\n", THREADID);

  ATOMIC(nr_threads_running++);

  thread_info[THREADID] = T_RUNNING;

  // start execution
  (*helper_thread_info->start_routine)(helper_thread_info->arg);

  fprintf(jiti_tty, "Thread %d ended.\n", THREADID);
  
  printf("---\nThread %d ended without calling pthread_exit()!\n", THREADID);
  dying(THREADID, 1);

  thread_info[THREADID] = T_EXITED;
  // end of thread
  free(helper_thread_info);
}      


///////////////////////////////////////


int dl_pthread_create(pthread_t * thread, const pthread_attr_t * attr,
                      void * (*start_routine)(void *), void * arg){  
  struct thread_info_t * thread_info;   
  
  thread_info=(struct thread_info_t *)malloc(sizeof(struct thread_info_t));   
  thread_info->start_routine  = start_routine+delta;
  thread_info->arg            = arg;           
  
  fprintf(jiti_tty, "Starting new thread...\n");

  ATOMIC(nr_threads_created++);

  return (*(int(*)(pthread_t*, const pthread_attr_t*, void*(*)(void*), void*))
	  reloc_info[PTHREAD_CREATE].real_address)
    (thread, attr, (void*(*)(void*))dl_pthread_create_helper-delta, thread_info);
}

 
///////////////////////////////////////


static stop(){
  fprintf(jiti_tty, "==== Program ended normally.\n");
  // hier ook testen op niet-vrijgegeven locks?
}


//////////////////////////////////////////


static start(){
  int i=0;

  fprintf(jiti_tty, "==== Deadlock detector: checking [%s]\n", jiti_program_name());
  
  if (!jiti_parallel_program()){
    fprintf(jiti_tty, "==== This is a sequential program => giving up.\n");
    return;
  }
  
  //// parallel program ////

  // onderschep einde
  jiti_callback("#STOP#", stop);
  
  // onderschep alle synchronisatiefuncties
  while (i<AANTAL_INSTR_FUNC){
    reloc_info[i].real_address = 
      jiti_callback(reloc_info[i].function, reloc_info[i].instr_address);
    i++;
  }
  
}


///////////////////////////////////////////


INIT(start);
