/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  instrument.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2002/02/28 12:24:13 $
 * $Log: instrument.c,v $
 * Revision 1.22  2002/02/28 12:24:13  ronsse
 * Merged updates from jmaebe branch, except the dynlib stuff.
 *
 * Revision 1.7.2.18  2002/02/27 17:40:29  jmaebe
 * * fixed some more entries in the opcode0F_table
 *
 * Revision 1.7.2.17  2002/02/27 15:03:14  jmaebe
 * * made RELOC and CLONE2INSTR thread-safe
 * * fixed use of UPDATE_RELOC macro
 * * moved LOOKUP_CLONE2INSTR from jiti.c to instrument.c
 * * fixed usage of LOOKUP_CLONE2INSTR (in most cases, an address in the
 *   original program instead of in the clone was passed, so it never found
 *   anything)
 *
 * Revision 1.7.2.16  2002/02/27 11:34:41  jmaebe
 * * fixed data races when instrumenting MT programs
 *
 * Revision 1.7.2.15  2002/02/27 10:19:25  jmaebe
 * * check if LOOKUP_RELOC() returns a value != 0 in loop() (in case someone
 *   uses loop in an unorthodox way and "loops" to an address ahead)
 * * decreased the INSTR_LEN's passed to no_mem in the opcode_0F table by 1
 *   since the 0F prefix is already counted in the ORIGINAL_INSTRUCTION now
 *   (this fixes a lot of problems!!)
 * + when using DEBUG, also add the address of the place where the return
 *   address from main is stored on the stack
 *
 * Revision 1.7.2.14  2002/02/25 15:29:23  jmaebe
 * * fixed some DEBUG code (activating DEBUG shoudln't crash jiti anymore)
 * + breakpoints corresponding to jmp/call_mem's in the clone are now
 *   replaced with jumps to X_instr-code if their length is >= 5 (this is the
 *   case for e.g. the jmp_mem's in the plt)
 *
 * Revision 1.7.2.13  2002/02/25 14:26:57  jmaebe
 * + jecxz support
 *
 * Revision 1.7.2.12  2002/02/22 17:03:42  jmaebe
 * * made checking of ret from main thread safe
 * * LOOKUP_RET_HANDLER now also uses LOOKUP_CLONE2INSTR and LOOKUP_RELOC when
 *   the ret address is in a dynlib
 * * ret_handler (handlers.s) now saves all registers (it calls the C-routine
 *   LOOKUP_RET_HANDLER), should be slightly faster and is now jumped at
 *   instead of called
 * * fac now works reliably too when instrumented!!
 *
 * Revision 1.7.2.11  2002/02/22 15:42:49  jmaebe
 * * fixed checking of ret from main. tt and "hello world" now work!!!!!!!
 *
 * Revision 1.7.2.10  2002/02/22 14:48:34  jmaebe
 * * fixed gs override support
 *
 * Revision 1.7.2.9  2002/02/22 14:23:31  jmaebe
 * + save return address of main at the start of the program and use it to
 *   stop instrumenting when we return from main
 *  
 * Revision 1.21  2002/02/21 13:14:20  ronsse
 * Added support for CMOVcc.
 *
 * Revision 1.7.2.8  2002/02/21 13:49:12  jmaebe
 * * both the loading of the value and the storing to the stack is now
 *   instrumented for "push mem"
 *
 * Revision 1.19  2002/02/20 12:05:08  jmaebe
 * * fixed bug in Poperand() (I didn't exclude the counting of the db 0x66
 *   from the instruction length passed to ORIGINAL_INSTRUCTION)
 *
 * Revision 1.18  2002/02/19 17:10:19  jmaebe
 * * cleaned up prefix counting (the only place they're still added to the
 *   instruction length, is in ORIGINAL_INSTRUCTION)
 * * changed a logical not to a binary not in btx_handler
 * - removed prefix_len parameter from expl_handler, use prefix_c instead
 *
 * Revision 1.17  2002/02/19 13:17:37  jmaebe
 * * all PUT_CHAR(X_instr,0x5n) changed to PUSH_REG() and POP_REG()
 * * fixed memory tracing in STACK_ACCESS for stores: when pushing, the
 *   memory at address -size(%esp) is accessed, not at (%esp)
 * + btx_handler: properly instruments memory accesses done by "btx %reg,mem"
 *   instructions (they have a range of 2^32 bits)
 * + support btr and bts
 *
 * Revision 1.16  2002/02/16 15:23:43  jmaebe
 * * fixed handling of 16bit versions of instructions with a 32bit constant
 *   argument (ie. cmpw 0x80,%ax) -> no_mem now also requires imm_len as one
 *   of its arguments
 *
 * Revision 1.15  2002/02/16 00:00:18  ronsse
 * MOV to/from segment registers now works
 *
 * Revision 1.14  2002/02/15 23:02:00  jmaebe
 * * STACK_ACCESS now has an extra parameter that specifies the size
 * * set X_kloon to 1 for interrupt_handler()
 * + pusha, popa, bt support implemented
 * + "pop mem" support implemented (used in setjmp of FPC)
 * + stosb, stosw, lodsb, lodsw support enabled
 *
 * Revision 1.13  2002/02/15 15:40:34  jmaebe
 * * re-added missing commit-messages
 *
 * Revision 1.12  2002/02/15 15:37:35  jmaebe
 * * fixed call_mem_handler (the length of the "call *mem" instructions was
 *   hard-coded to 6 for some dark reason :)
 *
 * Revision 1.11  2002/02/15 14:43:54  jmaebe
 * + support for "ret imm" (needed for Pascal code :)
 *
 * Revision 1.10  2002/02/15 13:12:07  jmaebe
 * + support for interrupts
 *
 * Revision 1.9  2002/02/14 17:37:07  jmaebe
 * * fixed bug with lock handling
 *
 * Revision 1.8  2002/02/14 15:42:19  jmaebe
 * + support for "lock"
 * + supportfor cmpxchg
 *
 * Revision 1.7  2002/02/07 13:19:52  jmaebe
 * * replaced occurrences of "PUT_CHAR(X_instr, 0x9C/0x9D)" with PUSHF/POPF
 * * slightly optimized jmp_mem_handler:
 *   * changed "pushf;addl $delta,%eax; popf" to "leal delta(%eax),%eax"
 *   * changed "pushl $0" (used to reserve room on the stack for the new
 *     target address) to "pushl %eax" (smaller machine code)
 *
 * Revision 1.6  2002/01/10 21:23:33  ronsse
 * *** empty log message ***
 *
 * Revision 1.5  2001/11/12 10:08:39  ronsse
 * *** empty log message ***
 *
 * Revision 1.4  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


//#define DEBUG

#include "jiti.h"

//////////////////////////////////////////////

// dummy argument voor arg1, arg2 en arg3
#define DUMMY 0

////// lock om de (parallelle) instrumentatie te beveiligen

//static pthread_mutex_t mmuutteexx;
//#define LOCK_DECODER()   pthread_mutex_lock(&mmuutteexx)
//#define UNLOCK_DECODER() pthread_mutex_unlock(&mmuutteexx)

volatile static int lloocckk = 0;
extern void __wait_and_set(unsigned);
#define LOCK_DECODER()   __wait_and_set((unsigned)&lloocckk)
#define UNLOCK_DECODER() lloocckk=0

void jiti_lock()   { LOCK_DECODER();   }
void jiti_unlock() { UNLOCK_DECODER(); }


//////

#define INSTR_LEN(X) X
#define IMM_LEN(X)   X
#define MEM_LEN(X)   X
#define CALC         0
#define TYPE(X)      X

extern int _jiti_par_exec;

// optimalisaties
unsigned FOLLOW_JMP_DIR = 1;
unsigned FOLLOW_CALL_DIR = 1;
unsigned MAX_FOLLOW = 10;

unsigned GEBRUIK_RELOC_INFO = 1;
#define RELOC 200

unsigned GEBRUIK_PATCH_INFO = 1;
#define PATCH 100

unsigned LOOKUP_RET = 1;

// instrumentaties
unsigned _instrument_code_ = 0;
unsigned _instrument_mem_ = 0;
unsigned _instrument_dyn_ = 0;
//

unsigned _X_start = 0;

void ret_handler_(void);
void jcc_taken_handler_(void);

unsigned follow = 0;

extern void trace_(void);
extern void trace_code_(void);

unsigned X_instr_start_block;

//// RELOCATIE INFO
volatile static int reloc_lock = 0;
#define LOCK_RELOC()   __wait_and_set((unsigned)&reloc_lock)
#define UNLOCK_RELOC() reloc_lock=0

unsigned reloc_i = 0;
struct _reloc_info_ {
	unsigned _from_;
	unsigned _to_;
} _reloc_info_[RELOC];

#ifdef USE_HASH
#	define HASH(key) (key%RELOC)
#	define UPDATE_RELOC(from, to) \
	if (GEBRUIK_RELOC_INFO) { \
		LOCK_RELOC(); \
		_reloc_info_[HASH((unsigned)from)]._from_ = from; \
		_reloc_info_[HASH((unsigned)from)]._to_   = to; \
		UNLOCK_RELOC(); \
	}
#else
#	define UPDATE_RELOC(from, to) \
	if (GEBRUIK_RELOC_INFO) { \
		LOCK_RELOC(); \
		_reloc_info_[reloc_i]._from_ = (unsigned)from; \
		_reloc_info_[reloc_i]._to_   = (unsigned)to;   \
		reloc_i=(reloc_i+1)%RELOC; \
		UNLOCK_RELOC(); \
	}
#endif

//// PATCH INFO
unsigned to_be_patched_i = 0;

struct _to_be_patched_info_ {
	unsigned _address_;
	unsigned _target_;
} _to_be_patched_info_[PATCH];

#define TO_BE_PATCHED(address, target) \
	if (GEBRUIK_PATCH_INFO) { \
		_to_be_patched_info_[to_be_patched_i]._address_ = (unsigned)address; \
		_to_be_patched_info_[to_be_patched_i]._target_  = (unsigned)target;   \
		to_be_patched_i=(to_be_patched_i+1)%PATCH;\
		DBG(fprintf(jiti_tty, "\n>>>[TO BE: <%x : %x>]", address, target));\
	}

unsigned X_address, X_len;	/* algemene info */
unsigned X_kloon;		/* is het ok om de instr in de kloon uit te voeren? */

//char X_instr_[64*1024*1024];

char *X_instr_ = 0;		// start v/d aangemaakte code
char *X_instr = 0;		// plaats waar nieuwe code geplaatst wordt

unsigned __instr_p__ = 0;

volatile static int clone2instr_lock = 0;
#define LOCK_CLONE2INSTR()   __wait_and_set((unsigned)&clone2instr_lock)
#define UNLOCK_CLONE2INSTR() clone2instr_lock=0
_clone2instr_t _clone2instr_[50000];

#define TRACE_CODE(from, to) \
	if (_instrument_code_) { \
		DBG(fprintf(jiti_tty, "\n>>>TRACE_CODE(%x, %x)", from, to)); \
		PUSHL32(X_instr, from);    X_instr+=5; \
		PUSHL32(X_instr, to);      X_instr+=5; \
		PUT_CALL(X_instr, trace_code_-(unsigned)X_instr-5);  X_instr+=5; \
	}

unsigned _fp(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _mmx(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _3dnow(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _sse(unsigned dummy1, unsigned dummy2, unsigned dummy3);

//unsigned char opcode1;
unsigned from_address;
unsigned to_address;

unsigned extended0F(unsigned, unsigned, unsigned);

//unsigned len;
unsigned address;

#include <stdio.h>
extern FILE *jiti_tty;

unsigned char byte1, byte2, byte3, byte4, prefix_c, immsize_override, lock, seg_override;

int jiti_fprint_ins1(FILE * file, unsigned printaddress, unsigned address);

/////////////////////

#define GIVE_INFO\
        jiti_fprint_ins1(jiti_tty, 1, address);\
        fprintf(jiti_tty, "\n%x %x %x %x\n", byte1, byte2, byte3, byte4)
  

unsigned todo(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction is not yet instrumented!");
	return 0;
}

unsigned dontexist(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction doesn't exist. Go away!");
	return 0;
}

unsigned unsupported(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction isn't supported. Go away!");
	return 0;
}

unsigned privileged(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This is a privileged instruction!.");
	return 0;
}

///////////////////////////////////////////////////////

void STACK_ACCESS(unsigned type, unsigned size){

	if (lock)
		panic("lock set in STACK_ACCESS??");

	DBG(fprintf(jiti_tty, "\n>>> STACK_ACCESS(%d)", type));
	if (_instrument_mem_) {
		// stack_access is called for push/pop/pusha/popa. When pushing (type ==
		// TYPE(STORE)), the value(s) is (are) stored between ESP-size and ESP
		// -> we need to pass ESP-size to trace! By adjusting ESP now (instead of
		// substracting size from the pushed ESP later), we can guarantee that
		// the callback can also access this memory in case it wants to
		if (type == TYPE(STORE)) {
			if (size > 127)
				panic("size > 127 in STACK_ACCESS!!");

			PUT_LONG(X_instr,(-(signed char)size) << 24 | 0x24648d);
			X_instr += 4;		// leal -size(%esp),%esp
		}
		
		PUSH_REG(ESP);               X_instr++;       // pushl %esp
		PUSHL8(X_instr, size);       X_instr += 2;		// pushl len
		PUSHL8(X_instr, type);       X_instr += 2;		// LOAD / STORE
		PUSHL32(X_instr, X_address); X_instr += 5;		// pushl X_address
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	
		if (type == TYPE(STORE)) {
			// restore ESP
			PUT_LONG(X_instr,size << 24 | 0x24648d);
			X_instr += 4;		// leal size(%esp),%esp
		}
	} else
		X_kloon = 1;
}

///////////////////

#define MAKE_LINK(from, to) \
	DBG(fprintf(jiti_tty,"\n>>>MAKE_LINK(%x, %x)", (from), (to)-(from)-5));\
	JMPL((from), (to)-(from)-5);

///////////////////

void ORIGINAL_INSTRUCTION(unsigned instr_len){
	
	instr_len += prefix_c;
	// clear lock (it's counted with prefix_c)
	lock = 0;

	if (instr_len >= 5) {
		//    fprintf(jiti_tty, "LEN>=5, X_instr_start_block=%x\n", X_instr_start_block); 
		MAKE_LINK(X_address + delta, X_instr_start_block);
	} else {
		if (X_kloon)
			memcpy((void *) (X_address + delta), (void *) X_address, instr_len);	// else { houdt BP; }
	}

	memcpy((void *) X_instr, (void *) X_address, instr_len);
	X_instr += instr_len;
	X_address += instr_len;
}

////////////////////

unsigned leave() {
	if (lock)
		panic("lock before leave??");
	PUT_CHAR(X_instr, 0x89);
	X_instr++;			/* movl %ebp, %esp */
	PUT_CHAR(X_instr, 0xEC);
	X_instr++;
	STACK_ACCESS(TYPE(LOAD),4);
	POP_REG(EBP);
	X_instr++;			/* popl %ebp */
	return X_address + 1;
}

////////////////////

unsigned nop(unsigned instr_len, unsigned dummy1, unsigned dummy2) {
	
	if (lock)
		panic("lock before nop??");

	*(unsigned char *) (X_address + delta) = 0x90;	/* nop plaatsen in kloon, en niets in instr */
	return X_address + instr_len;
}

////////////////////

unsigned movl_eax_mem_handler(unsigned instr_len, unsigned mem_len, unsigned mem_type){
	if (_instrument_mem_) {
		PUSHL32(X_instr, GET_LONG(X_address + 1 + prefix_c));
		X_instr += 5;
		PUSHL8(X_instr, mem_len);
		X_instr += 2;
		PUSHL8(X_instr, mem_type);
		X_instr += 2;
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	}
	ORIGINAL_INSTRUCTION(instr_len);
	return X_address;
}

////////////////////////////////////////////////////////////

unsigned cmovcc(unsigned dummy1, unsigned dummy2, unsigned dummy3){


  if (byte2>>6==3) return no_mem(INSTR_LEN(2),IMM_LEN(0), 0);

  // Transformation:
  //       CMOVcc mem, reg
  // becomes
  //       Jcc X
  //       JMP Y
  //    X: code generated by check_mod()
  //    Y: CMOVcc mem, reg

  if (_instrument_mem_){
    PUT_CHAR(X_instr, byte1 | 0x30); X_instr++;             // Jcc
    PUT_CHAR(X_instr, 2); X_instr++;                        // disp8
    PUT_CHAR(X_instr, 0xEB); X_instr++;                     // jmp
    PUT_CHAR(X_instr, 19+get_mem_len(byte2)); X_instr++;    // disp8
  }

  return check_mod(IMM_LEN(0), MEM_LEN(4), TYPE(LOAD));

}

////////////////////////////////////////////////////////////

unsigned push_pop_handler(unsigned instr_len, unsigned type, unsigned dummy1){
	STACK_ACCESS(type,4);
	ORIGINAL_INSTRUCTION(instr_len);
	return X_address;
}

////////////////////////////////////////////////////////////////

unsigned push_pop_mem_handler(unsigned stacktype, unsigned memtype, unsigned dummy3){

	STACK_ACCESS(TYPE(stacktype),4);
	return check_mod(IMM_LEN(0),MEM_LEN(4),TYPE(memtype));
}

////////////////////////////////////////////////////////////////

void expl_handler(unsigned leal_len, unsigned mem_len, unsigned mem_type){

	/* bewaar EAX */

	// we moeten een lokale variabele gebruiken voor par. progr.
	PUSH_REG(EAX);
	X_instr++;			// pushl %eax

	// hier moeten we nog testen of de adresexpressie SP gebruikt!!!! Indien ja, moet de
	// berekende waarde gewijzigd worden!!

	/* maak leal ..., %eax */
	memcpy((void *) X_instr, (void *) (X_address + prefix_c), leal_len);	/* kopieer */
	PUT_CHAR(X_instr, 0x8D);	/* lea */
	PUT_CHAR(X_instr + 1, GET_CHAR(X_instr + 1) & 0xC7);	/* , %eax */
	X_instr += leal_len;

	/* call trace_ */
	PUSH_REG(EAX);
	X_instr++;			/* pushl %eax */
	PUSHL8(X_instr, mem_len);
	X_instr += 2;
	PUSHL8(X_instr, mem_type);
	X_instr += 2;
	PUSHL32(X_instr, X_address);
	X_instr += 5;
	PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
	X_instr += 5;

	/* herstel EAX */
	POP_REG(EAX);
	X_instr++;			// popl %ax

	// MERK OP: geen originele instructie! 
}

////////////////////////////////////////////////////////////////////////////

unsigned get_mem_len(unsigned byte2__NOT_USED){
	unsigned char mod = byte2 >> 6;
	unsigned char rm = byte2 & 7;
	unsigned char base = byte3 & 7;

	//  printf("-B2=%x M%d RM%d B%d-", byte2, mod, rm, base);

	/* mod mag niet 3 zijn */
	if (mod == 3) panic("get_mem_len(reg1,reg2)");

	if ((mod == 0) && (rm != 4) && (rm != 5)) return 0;
	if ((mod == 0) && (rm == 5)) return 4;
	if ((mod == 1) && (rm != 4)) return 1;
	if ((mod == 2) && (rm != 4)) return 4;

	/* nu moet rm==4 zijn! => SIB */
	if (rm != 4) panic("get_mem_len: rm==4");

	if ((mod == 0) && (base != 5)) return 1;
	if ((mod == 0) && (base == 5)) return 5;
	if (mod == 1) return 2;
	if (mod == 2) return 5;		/* 5?? */

	/* hier mogen we nooit komen */
	panic("get_mem_len: not reached");
	return 0;
}

////////////////////////////////////////////////////////////////////////////

unsigned loop(unsigned dummy0, unsigned dummy1, unsigned dummy2){
	// loop* is 2-bytes lang en heeft een 8-bit displacement
	signed char disp8;
	unsigned target, new_target;

	if (lock)
		panic("lock before loop??");

	disp8 = GET_CHAR(X_address + 1);
	target = X_address + (signed) disp8 + 2;

	// try to find a better target
	if ((new_target = LOOKUP_RELOC(target)) != 0)
		target = new_target;

	// Change
	//    LOOP X
	// to
	//    LOOP Y   ; 2 bytes
	//    JMP Z    ; 5 bytes
	// Y: JMP X    ; 5 bytes
	// Z:

	memcpy(X_instr, (char *) X_address, 1);           X_instr++;
	PUT_CHAR(X_instr, 5);                             X_instr++;
	JMPL(X_instr, 5);                                 X_instr += 5;
	JMPL(X_instr, target - (unsigned) X_instr - 5);   X_instr += 5;
	return X_address + 2;

}

////////////////////////////////////////////////////////////////////////////

unsigned no_mem(unsigned length, unsigned imm_len, unsigned dummy2){
	X_kloon = 1;

	if (length == CALC) {	/* we moeten zelf nog de lengte bepalen, enkel gebruikt voor LEAL */
		length = 2 + get_mem_len(byte2);
	} else {
		if (imm_len > immsize_override) {
			length -= immsize_override;
		}
	}

	ORIGINAL_INSTRUCTION(length);

	return X_address;

}

////////////////////////////////////////////////////////////////////////////

unsigned check_mod(unsigned imm_len, unsigned mem_len, unsigned type){
	unsigned leal_len, instr_len;
	unsigned char mod2 = byte2 >> 6;

	if (mod2 != 3) {		/* memory operation */

		leal_len = get_mem_len(byte2);

		if (imm_len) {
			instr_len = leal_len + imm_len;
			if (imm_len > immsize_override)
				instr_len -= immsize_override;	// 4 wordt 2
		} else {
			instr_len = leal_len;
		}

		leal_len += 2;
		instr_len += 2;

		//printf("@0x%x [IMM%d INSTRLEN%d LEALLEN%d OVERRIDE%d MEMLEN%d]\n", X_address, imm_len, instr_len, leal_len, immsize_override, mem_len); ////

		if (_instrument_mem_)
			expl_handler(leal_len, mem_len, type);
		else
			X_kloon = 1;

		ORIGINAL_INSTRUCTION(instr_len);
		return X_address;

	} else {			/* reg,reg operation */
/*
		if ((byte1 == 0xc1) && immsize_override) {
			immsize_override = 0;
			printf("***********X_address=%x\n", X_address);
			return no_mem(4, DUMMY, DUMMY);	// 66 c1 e5 08 van GOGO
		}
*/
		return no_mem(INSTR_LEN(2 + imm_len), imm_len, DUMMY);
	}

}

//////////////////////////////////////////////////////////

unsigned check_mod_op2a(unsigned imm_len, unsigned mem_len, unsigned dummy){
	/* op2 testen voor L/X */
	unsigned type = 0;
	unsigned opcode2 = (byte2 & 0x3F) >> 3;

	switch (opcode2) {
		case 0:			/* add */
			type = TYPE(MODIFY);
			break;
		case 1:			/* or */
			type = TYPE(MODIFY);
			break;
		case 2:			/* adc */
			type = TYPE(MODIFY);
			break;
		case 3:			/* sbb */
			type = TYPE(MODIFY);
			break;
		case 4:			/* and */
			type = TYPE(MODIFY);
			break;
		case 5:			/* sub */
			type = TYPE(MODIFY);
			break;
		case 6:			/* xor */
			type = TYPE(MODIFY);
			break;
		case 7:			/* cmp */
			type = TYPE(LOAD);
			break;
		default:
			panic("check_mod_op2a: check type");
	}

	return check_mod(IMM_LEN(imm_len), MEM_LEN(mem_len), TYPE(type));
}

///////////////////////////////////////////////////////////

unsigned check_mod_op2b(unsigned imm_len, unsigned mem_len, unsigned type){
	/* op2 testen voor L/X */
	unsigned opcode2 = (byte2 & 0x3F) >> 3;

	switch (opcode2) {
		case 0:
			return check_mod(IMM_LEN(imm_len), MEM_LEN(mem_len), TYPE(LOAD));	// TEST
		case 2:
		case 3:
			return check_mod(IMM_LEN(0), MEM_LEN(mem_len), TYPE(MODIFY));	// NOT/NEG mem
		case 4:
		case 5:
		case 6:
		case 7:
			return check_mod(IMM_LEN(0), MEM_LEN(mem_len), TYPE(LOAD));	// MUL, IMUL, DIV, IDIV 
			// (E)A[XL] with mem
		default:
			printf("\n\n+++TODONOW: opcode2=%d+++\n\n", opcode2);
			_exit(-1);
	}
	// not reached
	return 0;

}

///////////////////////////////////////////////////////////

unsigned Poperand();
unsigned Plock();
unsigned PSegOverride(unsigned reg, unsigned dummy1, unsigned dummy2);

unsigned Paddress(){
	panic("--address size--");
	// not reached
	return 0;
}

#define REP_NONE  0
#define REP_LOAD  1
#define REP_STORE 2

extern unsigned range();

//////////////////////////////////////////////////////

unsigned interrupt_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	X_kloon = 1;

	/* trace code */
	TRACE_CODE(_X_start, X_address+1);

	/* replicate in clone and in X_instr block*/
	ORIGINAL_INSTRUCTION(2);

	/* jump to the next instruction */
	JMPL(X_instr,X_address + delta - (unsigned) X_instr - 5);
	X_instr+=5;
	/* don't instrument any further, this could be the end of the program */
	return 0;
}

//////////////////////////////////////////////////////

unsigned ret_handler(unsigned len, unsigned dummy2, unsigned dummy3){
	
	int retvalue;
	
	if (lock)
		panic("lock before ret??");

	/* trace mem */
	STACK_ACCESS(TYPE(LOAD),4);

	/* trace code */
	TRACE_CODE(_X_start, X_address+len-1);

	if (len == 3) {
		// ret imm
		retvalue = GET_SHORT(X_address+1);
		
		// save eax
		PUSH_REG(EAX);
		X_instr++;

		// load the return address
		PUT_LONG(X_instr,0x0424448b);
		X_instr+=4;	// movl  0x4(%esp,1),%eax

		/* stack lay-out now:
		 * esp												: saved eax
		 * esp+4											:	return address
		 * esp+8 till esp+8+retvalue	: arguments
		 */

		// save it at the start of the arguments
		if (retvalue+4 <= 127) {
			PUT_LONG(X_instr,(retvalue+4) << 24 | 0x244489);
			X_instr += 4;	// movl  %eax,imm8(%esp)
		} else {
			PUT_SHORT(X_instr,0x8489);
			PUT_CHAR(X_instr+2,0x24);
			PUT_LONG(X_instr+3,retvalue+4);
			X_instr += 7;	// movl %eax,imm32(%esp)
		}

		// restore eax
		POP_REG(EAX)
		X_instr++;
		
		// remove (old) return address and all arguments except for the first 4
		// bytes (where the return address is stored now)	
		if (retvalue <= 127) {
			// use short form
			PUT_LONG(X_instr,retvalue << 24 | 0x24648d);
			X_instr += 4;		// leal imm8(%esp),%esp
		} else {
			// use long form
			PUT_SHORT(X_instr,0xa48d);
			PUT_CHAR(X_instr+2,0x24);
			PUT_LONG(X_instr+3,retvalue);
			X_instr += 7;		// leal imm32(%esp),%esp
		}
	}

	JMPL(X_instr, ret_handler_ - (unsigned) X_instr - 5);
	X_instr += 5;

	return 0;			// einde van dit basisblok
}

//////////////////////////////////////////////////////

unsigned jmp_dir_handler(unsigned len, unsigned dummy1, unsigned dummy2){
	unsigned new_target, target;
	signed char disp8;
	extern unsigned plt_end;

	if (lock)
		panic("lock before jmp_dir??");

	if (len == 2) {		/* disp8  */
		disp8 = (signed) GET_CHAR(X_address + 1);
		target = disp8 + X_address + 2;
	} else {			/* disp32 */
		target = GET_LONG(X_address + 1) + X_address + 5;
		MAKE_LINK(X_address + delta, X_instr);
	}

	if (target < plt_end) {
		printf("XXXX JMP_DIR_HANDLER met TARGET=0x%x<PLT_END XXXX\n", target);
		JMPL(X_instr, target - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	}

	if (GEBRUIK_RELOC_INFO) {
		if ((new_target = LOOKUP_RELOC(target))) {
			DBG(fprintf(jiti_tty, "\n>>>FOUND @ %x", new_target));
			//TRACE_CODE(_X_start, X_address+len-1); _X_start=target;
			JMPL(X_instr, new_target - (unsigned) X_instr - 5);
			X_instr += 5;
			return 0;
		}
	}

	TRACE_CODE(_X_start, X_address + len - 1);
	_X_start = target;

	if (FOLLOW_JMP_DIR) {
		JMPL(X_instr, target + delta - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	} else {
		if (follow++ > MAX_FOLLOW) {
			JMPL(X_instr, target + delta - (unsigned) X_instr - 5);
			X_instr += 5;
			return 0;
		} else {
			DBG(fprintf(jiti_tty, "\n>>>Restart from %x", target));
			return target;
		}
	}

}

//////////////////////////////////////////////////////

void jmp_reg_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before jmp_reg??");

	TRACE_CODE(_X_start, X_address + 1);

	PUT_CHAR(X_instr, (GET_CHAR(X_address + 1) & 7) | 0x50);
	X_instr++;			/* PUSHL %reg */
	PUSHF;
	X_instr++;			/* pushf */
	ADDL_4ESP(X_instr, delta);
	X_instr += 8;		/* addl $delta, 4(%esp) */
	POPF;
	X_instr++;			/* popf */
	PUT_CHAR(X_instr, 0xC3);
	X_instr++;			/* ret */
}

//////////////////////////////////////////////////////

void jmp_mem_handler(unsigned len, unsigned dummy1, unsigned dummy2){

	if (lock)
		panic("lock before jmp_mem!!!");

	PUSH_REG(EAX);  // reserve space on the stack for the new target address
	X_instr++;			// pushl %eax (smaller than "pushl $0")
	PUSH_REG(EAX);
	X_instr++;			// pushl %eax (it will be used to calculate the new target)

	TRACE_CODE(_X_start, X_address + len - 1);

	// load address of target in eax using a lea instruction
	memcpy((void *) X_instr, (void *) X_address, len);	// copy address expression
	PUT_CHAR(X_instr, 0x8D);	// lea
	PUT_CHAR(X_instr + 1, GET_CHAR(X_instr + 1) & 0xC7);	// ,%eax
	X_instr += len;

	/* trace eax */
	if (_instrument_mem_) {
		PUSH_REG(EAX);
		X_instr++;
		PUSHL8(X_instr, 4);
		X_instr += 2;
		PUSHL8(X_instr, TYPE(LOAD));
		X_instr += 2;
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	}

	/* get target address */
	PUT_CHAR(X_instr, 0x8B);
	X_instr++;			// movl (%eax),%eax
	PUT_CHAR(X_instr, 0x00);
	X_instr++;

	/* adjust target address so it points to the clone          */
	/* (use lea instead of add so the flags don't get modified) */
	PUT_CHAR(X_instr,0x8d);
	X_instr++;      // leal delta(%eax),%eax
	PUT_CHAR(X_instr,0x80);
	X_instr++; 
	PUT_LONG(X_instr,delta);
	X_instr+=4; 

	PUT_LONG(X_instr, 0x04244489);
	X_instr += 4;		// movl %eax, 4(%esp)
	POP_REG(EAX);
	X_instr++;
	RET;
	X_instr++;
}

//////////////////////////////////////////////////////

unsigned call_dir_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	unsigned new_target;
	extern unsigned plt_end;
	unsigned target = GET_LONG(X_address + 1) + X_address + 5;

	if (lock)
		panic("lock before call_dir??");

	TRACE_CODE(_X_start, X_address + 4);

	X_len = 5;

	/* link leggen van kloon naar instr */
	MAKE_LINK(X_address + delta, X_instr);
	/* trace */
	STACK_ACCESS(TYPE(STORE),4);

	DBG(fprintf
			(jiti_tty, "\n>>>0x%x: CALL to 0x%x: %c PLT", X_address, target,
			 target < plt_end ? '<' : '>'));

	if (target < plt_end) {	/* dyn gelinkte functie */

#if 1
		// returnadres in de kloon, we wijzigen "ret" namelijk niet in de dyn. lib.
		PUSHL32(X_instr, X_address + 5 + delta);
		X_instr += 5;
		// we springen naar de dyn. lib.
		JMPL(X_instr, target - (unsigned) X_instr - 5);
		X_instr += 5;
		// instr. is gedaan
		return 0;
#else
		// gewoon de call uitvoeren ...
		PUT_CALL(X_instr, target - (unsigned) X_instr - 5);
		X_instr += 5;
		// ... en verder instrumenteren
		_X_start = X_address + 5;
		return X_address + 5;
#endif

	} else {			// statisch gelinkte functie

		// returnadres wijst naar de ori. code, we wijzigen de bijhorende "ret" namelijk.
		PUSHL32(X_instr, X_address + 5);
		X_instr += 5;
		if (GEBRUIK_RELOC_INFO) {
			if ((new_target = LOOKUP_RELOC(target))) {
				DBG(fprintf
						(jiti_tty, "\n>>>Found target in reloc info! -> %x",
						 new_target));
				JMPL(X_instr, new_target - (unsigned) X_instr - 5);
				X_instr += 5;
				return 0;
			} else if ((new_target = LOOKUP_CLONE2INSTR(target))) {
				DBG(fprintf
						(jiti_tty,
						 "\n>>>Found target in clone2instr info! -> %x",
						 new_target));
				JMPL(X_instr, new_target - (unsigned) X_instr - 5);
				X_instr += 5;
				return 0;
			}
		}

		if (FOLLOW_CALL_DIR) {
			JMPL(X_instr, target + delta - (unsigned) X_instr - 5);
			X_instr += 5;
			return 0;
		} else {
			if (follow++ > MAX_FOLLOW) {
				JMPL(X_instr, target + delta - (unsigned) X_instr - 5);
				X_instr += 5;
				return 0;
			} else {
				DBG(fprintf(jiti_tty, "\n>>>Restart from %x", target));
				_X_start = target;
				return target;
			}
		}
	}
}

//////////////////////////////////////////////////////

void call_reg_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before call_reg??");

	TRACE_CODE(_X_start, X_address + 1);

	/* terugkeeradres */
	PUSHL32(X_instr, X_address + 2);
	X_instr += 5;
#ifdef MEM
	/* nog tracen! */
#endif
	jmp_reg_handler(DUMMY, DUMMY, DUMMY);

}

//////////////////////////////////////////////////////

void call_mem_handler(unsigned len, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before call_mem!!");

	TRACE_CODE(_X_start, X_address);

	/* terugkeeradres */
	PUSHL32(X_instr, X_address + len);
	X_instr += 5;

#ifdef MEM
	/* nog tracen! */
#endif

	jmp_mem_handler(len, DUMMY, DUMMY);
}

//////////////////////////////////////////////////////

unsigned jcc_taken_handler(unsigned address){
	unsigned new_target, jcc_instr =
		*(unsigned *) (address), patch_address = address - 5;
	unsigned target;
	signed char disp8;

	// bereken target & condcode
	if (*(char *) jcc_instr != 0x0F) {	// disp8
		disp8 = GET_CHAR(jcc_instr + 1);
		target = jcc_instr + (signed) disp8 + 2;
	} else {			// disp32
		target = jcc_instr + GET_LONG(jcc_instr + 2) + 6;
	}

	DBG(fprintf(jiti_tty, "\n>>>In jcc_taken_handler(), coming from %x, "
				"ori jcc @ %x, target=%x, patch address=%X",
				address, jcc_instr, target, patch_address));


	new_target = LOOKUP_CLONE2INSTR(target+delta);

	if (GEBRUIK_RELOC_INFO)
		if (!new_target)
			new_target = LOOKUP_RELOC(target);

	// decodeer
	if (!new_target)
		new_target = decode_block(target + delta);
	// wijzig instr
	LOCK_DECODER();
	MAKE_LINK(patch_address, new_target);
	UNLOCK_DECODER();

	DBG(jiti_fprint_ins1(jiti_tty, 1, patch_address));

	DBG(fprintf
			(jiti_tty, "\n>>>DONE (jcc_taken_handler), restarting at %x",
			 new_target));

	FLUSH_INSTRUCTIONS;

	// ok
	return new_target;
}

//////////////////////////////////////////////////////

unsigned jcc_handler(unsigned len, unsigned dummy1, unsigned dummy2){
	unsigned target, new_target, condcode, jecxz;
	signed char disp8;

	if (lock)
		panic("lock before jcc??");

	jecxz = FALSE;
	// bereken target & condcode
	if (len == 2) {
	
		disp8 = GET_CHAR(X_address + 1);
		target = X_address + (signed) disp8 + 2;
		condcode = GET_CHAR(X_address) & 0x0F;
		memcpy((char *) (X_address + delta), (char *) X_address, 2);
	} else {
		target = X_address + GET_LONG(X_address + 2) + 6;
		condcode = GET_CHAR(X_address + 1) & 0x0F;
		MAKE_LINK(X_address + delta, X_instr);
	}

	TRACE_CODE(_X_start, X_address + len - 1);
	_X_start = X_address + len;

	DBG(fprintf(jiti_tty, "\n>>>TARGET=%x/condcode=%d/", target, condcode));

	if (jecxz) {
		PUSHF;
		X_instr++;
		PUT_SHORT(X_instr,0xc985);
		X_instr+=2;	// testl %ecx,%ecx 
	}

	if ((GEBRUIK_RELOC_INFO) && (jecxz == FALSE)) {
		new_target = LOOKUP_RELOC(target);
		if (new_target) {
			target = new_target;
			PUT_CHAR(X_instr, 0x0F);
			X_instr++;
			PUT_CHAR(X_instr, 0x80 | condcode);
			X_instr++;
			PUT_LONG(X_instr, target - (unsigned) X_instr - 4);
			X_instr += 4;
			return X_address + len;
		} else {
			if (GEBRUIK_PATCH_INFO)
				TO_BE_PATCHED(X_instr + 2, target);
			target += delta;
		}
		DBG(fprintf(jiti_tty, "\n>>>newTARGET=%x ", target));
	} else {
		target += delta;
	}

#if 0
	PUT_CHAR(X_instr, 0x0F);
	X_instr++;
	PUT_CHAR(X_instr, 0x80 | condcode);
	X_instr++;
	PUT_LONG(X_instr, target - (unsigned) X_instr - 4);
	X_instr += 4;
#else
	target = (unsigned) jcc_taken_handler_;
	DBG(fprintf(jiti_tty, "\n>>>LINKING TO jcc_taken_handler() at %x!",
			 jcc_taken_handler_));

	PUT_CHAR(X_instr, 0x0F);
	X_instr++;			// Jcc +6+ord(jecxz)
	PUT_CHAR(X_instr, 0x80 | condcode);
	X_instr++;
	PUT_LONG(X_instr, 2+jecxz);
	X_instr += 4;

	if (jecxz == TRUE) {
		POPF;
		X_instr++;
	}
	PUT_CHAR(X_instr, 0xEB);
	X_instr++;			// JMP +9+ord(jecxz)
	PUT_CHAR(X_instr, 9+jecxz);
	X_instr++;

	if (jecxz == TRUE) {
		POPF;
		X_instr++;
	}
	PUT_CALL(X_instr, jcc_taken_handler_ - (unsigned) X_instr - 5);
	X_instr += 5;		// CALL jcc_taken_handler

	PUT_LONG(X_instr, X_address);
	X_instr += 4;		// adres van ori. jcc
#endif

	return X_address + len;

}

//////////////////////////////////////////////////////

unsigned pusha_popa_handler(unsigned type, unsigned dummy1, unsigned dummy2) {

	STACK_ACCESS(type,32);
	ORIGINAL_INSTRUCTION(1);
	return X_address;
}

unsigned btx_handler(unsigned type, unsigned dummy1, unsigned dummy2) {

	unsigned memlen;
	char srcreg, calcreg;

	if (byte2 >> 6 == 3) {
		// INSTR_LEN(2) because the 0x0f (and maybe other) prefix(es) will be
		// added by no_mem/ORIGINAL_INSTRUCTION
		return no_mem(INSTR_LEN(2),IMM_LEN(0),0);
	}

	memlen = get_mem_len(byte2);

	/* "btx reg, mem" can address a bit position within the range -2^31 to 
	 * 2^31 - 1 (bits) of mem -> we have to calculate the offset every
	 * time if we are tracing memory accesses
	 */
	if (_instrument_mem_) {

		if (immsize_override)
			panic("TODO: size override support for btx!!");

		srcreg = (GET_CHAR(X_address+prefix_c+2) & ~0xc7) >> 3;

		// we need two different registers
		if (srcreg != EAX)
			calcreg = EAX;
		else
			calcreg = EDX;

		PUSHF;
		X_instr++;       // save flags

		PUSH_REG(srcreg);
		X_instr++;
		PUSH_REG(calcreg);
		X_instr++;       // save registers used for address calculation

		// load address of target in calcreg using a lea instruction
		// note: if the memory expression uses esp, the calculated value should be
		// adjusted!!
		PUT_CHAR(X_instr, 0x8D);	// lea
		memcpy((void *) X_instr+1, (void *) X_address+prefix_c+1, memlen+1);	// copy address expression
		PUT_CHAR(X_instr + 1, (GET_CHAR(X_instr + 1) & 0xC7) | (calcreg << 3));	// ,%calcreg
		X_instr+=memlen+2;

		// sar by 3 to get the byte in which the bit is modified
		// note: this is not a normal division by 8: -1 sar 3 = -1, but this is what
		//       we want: bit -1 relative to address addr lies in the byte at addr-1

		PUT_CHAR(X_instr, 0xc1);
		X_instr++;                                       // sarl
		PUT_CHAR(X_instr, (0x1f << 3) | srcreg);
		X_instr++;                                       // , %srcreg
		PUT_CHAR(X_instr, 3);
		X_instr++;                                       // $3

		// mask out the two lower bits, since the processor will access the int
		// at address+(4*(bitoffset div 32)) (page 3-45 of the "Intel Architecture
		// Software Developers Manual, Volume 2: Instruction Set Reference"

		PUT_CHAR(X_instr,0x83);
		X_instr++;
		PUT_CHAR(X_instr,0xe0 | srcreg);
		X_instr++;
		PUT_CHAR(X_instr,~3);
		X_instr++;                                       // andl $~3,%srcreg

		// add it to the address
		PUT_CHAR(X_instr,0x1);                           // add
		X_instr++;
		PUT_CHAR(X_instr,0xc0 | (srcreg << 3) | calcreg);
		X_instr++;                                       // %srcreg,%calcreg

		PUSH_REG(calcreg);
		X_instr++;  // push the resulting address on the stack

		PUSHL8(X_instr,4);
		X_instr += 2;  // this instruction accesses 4 bytes, see above

		PUSHL8(X_instr,type)
		X_instr += 2;  // the kind of access

		PUSHL32(X_instr,X_address);
		X_instr += 5; // the address of the access

		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5; // trace

		// restore saved registers and flags
		POP_REG(calcreg);
		X_instr++;

		POP_REG(srcreg);
		X_instr++;

		POPF;
		X_instr++;

	} else {
	  X_kloon = 1;
	}

	ORIGINAL_INSTRUCTION(2+memlen);
	return X_address;
}

//////////////////////////////////////////////////////

void rep_handler(unsigned si_type, unsigned di_type){

	/* voor de instructie */
	if (di_type != REP_NONE) {
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUSH_REG(EDI);
		X_instr++;
	}

	if (si_type != REP_NONE) {
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUSH_REG(ESI);
		X_instr++;
	}

	/* de instructie zelf */
	memcpy((void *) X_instr, (void *) X_address, X_len);
	X_instr += X_len;

	/* na de instructie */
	if (si_type == REP_LOAD) {	/* REP_STORE kan niet! */
		PUSH_REG(ESI);
		X_instr++;
		PUSHL8(X_instr, TYPE(LOAD));
		X_instr += 2;
		PUT_CALL(X_instr, range - (unsigned) X_instr - 5);
		X_instr += 5;
	}

	if (di_type != REP_NONE) {
		PUSH_REG(EDI);
		X_instr++;
		PUSHL8(X_instr, di_type == REP_LOAD ? TYPE(LOAD) : TYPE(STORE));
		X_instr += 2;
		PUT_CALL(X_instr, range - (unsigned) X_instr - 5);
		X_instr += 5;
	}

}

///////////////////////////////////////////////////////

unsigned rep(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	unsigned si_type = 0, di_type = 0;

	if (lock)
		panic("lock before rep??");

	X_len = 1;

	if ((byte1 == 0xF2) || (byte1 == 0xF3)) {	// REPNE of REP/REPE
		X_len++;
		byte1 = byte2;
		byte2 = byte3;
		byte3 = byte4;
	}

	if (byte1 == 0x66) {	// OPERAND SIZE PREFIX
		X_len++;
		byte1 = byte2;
		byte2 = byte3;
		byte3 = 0;
	}

	switch (byte1) {
		case 0xA6:			/* REPNE CMPSb */
			si_type = REP_LOAD, di_type = REP_LOAD;
			break;
		case 0xA7:			/* REPNE CMPSw */
			si_type = REP_LOAD, di_type = REP_LOAD;
			break;
		case 0xAE:			/* REPNE SCASb */
			si_type = REP_NONE, di_type = REP_LOAD;
			break;
		case 0xAF:			/* REPNE SCASw */
			si_type = REP_NONE, di_type = REP_LOAD;
			break;
		case 0x6C:			/* REP INSb    */
			panic("REP PORT!");
		case 0x6D:			/* REP INSw    */
			panic("REP PORT!");
		case 0xAC:			/* REP LODSb   */
			si_type = REP_LOAD, di_type = REP_NONE;
			break;
		case 0xAD:			/* REP LODSw   */
			si_type = REP_LOAD, di_type = REP_NONE;
			break;
		case 0xA4:			/* REP MOVSb   */
			si_type = REP_LOAD, di_type = REP_STORE;
			break;
		case 0xA5:			/* REP MOVSw   */
			si_type = REP_LOAD, di_type = REP_STORE;
			break;
		case 0x6E:			/* REP OUTSb   */
			panic("REP PORT!");
		case 0x6F:			/* REP OUTSw   */
			panic("REP PORT!");
		case 0xAA:			/* REP STOSb   */
			si_type = REP_NONE, di_type = REP_STORE;
			break;
		case 0xAB:			/* REP STOSw   */
			si_type = REP_NONE, di_type = REP_STORE;
			break;
		default:
			panic("Unknown rep!");
	}
	if (0 /*_instrument_mem_*/ ) {
		rep_handler(si_type, di_type);
		return X_address + X_len;
	} else {
		X_kloon = 1;
		ORIGINAL_INSTRUCTION(X_len);
		return X_address;
	}

}

/////////////////////////////////////////////////////////////

unsigned opcodeFF(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	unsigned char opcode2 = (byte2 & 0x3F) >> 3;
	unsigned char mod2 = byte2 >> 6;

	switch (opcode2) {
		case 0:		/* DEC of INC MEM */
		case 1:		/* DEC of INC MEM */
			return check_mod(IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY));
		case 2:
		case 3:
			if (mod2 == 3) {	/* CALL *REG */
				call_reg_handler(DUMMY, DUMMY, DUMMY);
				return 0;
			} else {		/* CALL *MEM */
				call_mem_handler(2 + get_mem_len(byte2), DUMMY, DUMMY);
				return 0;
			}
		case 4:
		case 5:
			if (mod2 == 3) {	/* JMP *REG */
				jmp_reg_handler(DUMMY, DUMMY, DUMMY);
				return 0;
			} else {		/* JMP *MEM */
				jmp_mem_handler(2 + get_mem_len(byte2), DUMMY, DUMMY);
				return 0;
			}
		case 6:			/* PUSH *MEM */
			return push_pop_mem_handler(TYPE(STORE), TYPE(LOAD), DUMMY);
		case 7:
			return dontexist(DUMMY, DUMMY, DUMMY);	/* illegal */
	}
	panic("assert");
	return -1;			// not reached
}

//////////////////////////////////////////////////////////////////////////

typedef struct decode {
	unsigned (*function) (unsigned, unsigned, unsigned);
	unsigned arg1;
	unsigned arg2;
	unsigned arg3;
} decode;

///////////////////////////////////////////////////////////////////////////

/* TABELLEN */

decode opcode_table[256] = {
	/* 0x00 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)}, /* ADD */
	/* 0x01 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)}, /* ADD */
	/* 0x02 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* ADD */
	/* 0x03 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* ADD */

	/* 0x04 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* ADD IMM8,A? */
	/* 0x05 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* ADD IMM32,A? */

	/* 0x06 */ {dontexist},
	/* 0x07 */ {dontexist},

	/* 0x08 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* OR */
	/* 0x09 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* OR */
	/* 0x0a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* OR */
	/* 0x0b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* OR */

	/* 0x0c */ {no_mem, INSTR_LEN(2),IMM_LEN(1)}, /* OR IMM,AL */
	/* 0x0d */ {no_mem, INSTR_LEN(5),IMM_LEN(4)}, /* OR IMM,EAX */

	/* 0x0e */ {dontexist},

	/* 0x0f */ {extended0F}, /* Jcc, MMX, SSE */

	/* 0x10 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* ADC */
	/* 0x11 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* ADC */
	/* 0x12 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* ADC */
	/* 0x13 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* ADC */

	/* 0x14 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* ADC IMM8, AL */
	/* 0x15 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* ADC IMM32, AX */

	/* 0x16 */ {dontexist},
	/* 0x17 */ {dontexist},

	/* 0x18 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* SBB */
	/* 0x19 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* SBB */
	/* 0x1a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* SBB */
	/* 0x1b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* SBB */

	/* 0x1c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* SBB IMM8, AL */
	/* 0x1d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* SBB IMM32, AL */

	/* 0x1e */ {dontexist},
	/* 0x1f */ {dontexist},

	/* 0x20 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* AND */
	/* 0x21 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* AND */
	/* 0x22 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* AND */
	/* 0x23 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* AND */

	/* 0x24 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* AND IMM8, AL */
	/* 0x25 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* AND IMM32, AL */

	/* 0x26 */ {unsupported}, /* ES override, not used in Linux */

	/* 0x27 */ {dontexist},

	/* 0x28 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* SUB */
	/* 0x29 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* SUB */
	/* 0x2a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* SUB */
	/* 0x2b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* SUB */

	/* 0x2c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* SUB IMM8,AL */
	/* 0x2d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* SUN IMM32,AL */

	/* 0x2e */ {unsupported}, /* CS override, not used in Linux */

	/* 0x2f */ {dontexist},

	/* 0x30 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* XOR */
	/* 0x31 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* XOR */
	/* 0x32 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* XOR */
	/* 0x33 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* XOR */

	/* 0x34 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* XOR IMM8,AL */
	/* 0x35 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* XOR IMM32,EAX */

	/* 0x36 */ {unsupported}, /* SS override, not used in Linux */

	/* 0x37 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* AAA */

	/* 0x38 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* CMP mem8,r  of CMP r1,r2 */
	/* 0x39 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* CMP mem32,r of CMP r1,r2 */
	/* 0x3a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* CMP r,mem8  of CMP r2,r1 */
	/* 0x3b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* CMP r,mem32 of CMP r2,r1 */

	/* 0x3c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* CMP IMM8, AL */
	/* 0x3d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* CMP IMM32, AL */

	/* 0x3e */ {unsupported}, /* DS override, not used in Linux */

	/* 0x3f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* AAS */

	/* 0x40 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x41 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x42 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x43 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x44 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x45 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x46 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x47 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */

	/* 0x48 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x49 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4a */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4b */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4c */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */

	/* 0x50 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EAX, 1 byte */
	/* 0x51 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ECX */
	/* 0x52 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EDX */
	/* 0x53 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EBX, SSE?? [TODO] */
	/* 0x54 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ESP */
	/* 0x55 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EBP */
	/* 0x56 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ESI */
	/* 0x57 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EDI */
	/* 0x58 */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EAX */
	/* 0x59 */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  ECX */
	/* 0x5a */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EDX */
	/* 0x5b */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EBX */
	/* 0x5c */ {dontexist}, /* POPL  ESP */
	/* 0x5d */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EBP */
	/* 0x5e */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  ESI */
	/* 0x5f */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EDI */

	/* 0x60 */ {pusha_popa_handler,TYPE(STORE)}, /* pusha */
	/* 0x61 */ {pusha_popa_handler,TYPE(LOAD)},  /* popa */
	/* 0x62 */ {dontexist},

	/* 0x63 */ {privileged}, /* ARPL */

	/* 0x64 */ {unsupported}, /* FS override, not used in Linux */
	/* 0x65 */ {PSegOverride, GS}, /* GS override, used in libpthread?? */

	/* 0x66 */ {Poperand}, /* prefix: operand size override */
	/* 0x67 */ {Paddress}, /* prefix: address size override */

	/* 0x68 */ {push_pop_handler, INSTR_LEN(5), TYPE(STORE)}, /* PUSHL IMM32, 5 bytes */

	/* 0x69 */ {check_mod, IMM_LEN(4), MEM_LEN(4), TYPE(LOAD)}, /* imull reg1,simm8,reg2 */

	/* 0x6a */ {push_pop_handler, INSTR_LEN(2), TYPE(STORE)}, /* PUSHL IMM8, 2 bytes */

	/* 0x6b */ {check_mod, IMM_LEN(1), MEM_LEN(1), TYPE(LOAD)}, /* imull reg1,imm32,reg2 */
	
	/* 0x6c */ {dontexist},
	/* 0x6d */ {dontexist},

	/* 0x6e */ {privileged}, /* OUTS */
	/* 0x6f */ {privileged}, /* OUTS */

	/* 0x70 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x71 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x72 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x73 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x74 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x75 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x76 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x77 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x78 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x79 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7a */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7b */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7c */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7d */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7e */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7f */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */

	/* 0x80 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(1)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met  */
	/* 0x81 */ {check_mod_op2a, IMM_LEN(4), MEM_LEN(4)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met w */
	/* 0x82 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(1)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met s */
	/* 0x83 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(4)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met sw */

	/* 0x84 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* TEST */
	/* 0x85 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* TEST */

	/* 0x86 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)}, /* XCHG */
	/* 0x87 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)}, /* XCHG */

	/* 0x88 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* MOV r1,r2 of MOV r,mem  8b */
	/* 0x89 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* MOV r1,r2 of MOV r,mem 32b */
	/* 0x8a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* MOV r2,r1 of MOV mem,r  8b */
	/* 0x8b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* MOV r2,r1 of MOV mem,r 32b */

	/* 0x8c */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* MOV ?S,r/m16 */

	/* 0x8d */ {no_mem, INSTR_LEN(CALC), IMM_LEN(0)}, /* LEAL */

	/* 0x8e */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* MOV r/m16,?S */

	/* 0x8f */ {push_pop_mem_handler, TYPE(LOAD), TYPE(STORE)}, /* pop mem, used in setjmp procedure of FPC */

	/* 0x90 */ {nop, INSTR_LEN(1)}, /* nop */

	/* 0x91 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x92 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x93 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x94 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x95 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x96 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x97 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */

	/* 0x98 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CBW */
	/* 0x99 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CDQ */

	/* 0x9a */ {dontexist},

	/* 0x9b */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* FWAIT */

	/* 0x9c */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* pushf: TODO */
	/* 0x9d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* popf:  TODO */

	/* 0x9e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* SAHF */

	/* 0x9f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, // TODO: LAHF

	/* 0xa0 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(1), TYPE(LOAD)}, /* MOV MEM, AL */
	/* 0xa1 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(4), TYPE(LOAD)}, /* MOV MEM, EAX */
	/* 0xa2 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(1), TYPE(STORE)}, /* MOV AL,  MEM */
	/* 0xa3 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(4), TYPE(STORE)}, /* MOV EAX, MEM */

	/* 0xa4 */ {rep}, /* movsb */
	/* 0xa5 */ {rep}, /* movsw */
	/* 0xa6 */ {rep /*restart, cmpsb_handler */ },	/* cmpsb */
	/* 0xa7 */ {rep}, /* cmpsw */

	/* 0xa8 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* TEST IMM8,AL */
	/* 0xa9 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* TEST IMM32, EAX */

	/* 0xaa */ {rep}, /* stosb */
	/* 0xab */ {rep}, /* stosw */
	/* 0xac */ {rep}, /* lodsb */
	/* 0xad */ {rep}, /* lodsw */
	/* 0xae */ {rep}, /* scasb */
	/* 0xaf */ {todo}, /* scasw */

	/* 0xb0 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb1 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb2 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb3 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb4 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb5 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb6 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb8 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb9 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xb7 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xba */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbb */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbc */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbd */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbe */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbf */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */

	/* 0xc0 */ {check_mod, IMM_LEN(1), MEM_LEN(1)}, /* RCL/RCR/ROL/ROR/SAR IMM */
	/* 0xc1 */ {check_mod, IMM_LEN(1), MEM_LEN(4)}, /* TODO */
	
	/* 0xc2 */ {ret_handler, INSTR_LEN(3)}, /* ret imm, 3 bytes */

	/* 0xc3 */ {ret_handler, INSTR_LEN(1)},	/* ret, 1 byte */

	/* 0xc4 */ {dontexist},
	/* 0xc5 */ {dontexist},

	/* 0xc6 */ {check_mod, IMM_LEN(1), MEM_LEN(1), TYPE(STORE)}, /* MOV IMM,REG */
	/* 0xc7 */ {check_mod, IMM_LEN(4), MEM_LEN(4), TYPE(STORE)}, /* MOV IMM,REG */

	/* 0xc8 */ {dontexist}, /* enter */
	/* 0xc9 */ {leave /*restart, leave_handler */ },	/* leave */

	/* 0xca */ {todo}, /* ret imm, other seg */
	/* 0xcb */ {todo}, /* ret, other seg */

	/* 0xcc */ {dontexist},
	/* 0xcd */ {interrupt_handler},
	/* 0xce */ {dontexist},
	/* 0xcf */ {dontexist},

	/* 0xd0 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* RCR/RCL/ROL/ROR/SAR: 1 */
	/* 0xd1 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)},
	/* 0xd2 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* : CL */
	/* 0xd3 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* : CL */

	/* 0xd4 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)}, /* AAM */
	/* 0xd5 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)}, /* AAD */

	/* 0xd6 */ {dontexist},
	/* 0xd7 */ {todo}, /* XLAT */

	/* FLOATING POINT START **************** */
	/* 0xd8 */ {_fp}, /* altijd 32bit toegang */
	/* 0xd9 */ {_fp}, /* altijd 32bit toegang */
	/* 0xda */ {_fp}, /* altijd 32bit toegang */
	/* 0xdb */ {_fp}, /* altijd 32bit of 80bit toegang */
	/* 0xdc */ {_fp}, /* altijd 64bit toegang */
	/* 0xdd */ {_fp}, /* altijd 64bit toegang */
	/* 0xde */ {_fp}, /* altijd 16bit toegang */
	/* 0xdf */ {_fp}, /* altijd 16bit of 64bit toegang */
	/* FLOATING POINT STOP **************** */

	/* 0xe0 */ {loop}, /* LOOPNZ */
	/* 0xe1 */ {loop}, /* LOOPZ */
	/* 0xe2 */ {loop}, /* LOOP */

	/* 0xe3 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0)},
	/* 0xe4 */ {no_mem,INSTR_LEN(2)}, /* IN PORT imm, reg8 */
	/* 0xe5 */ {no_mem,INSTR_LEN(2)}, /* IN PORT imm, reg32 */

	/* 0xe6 */ {dontexist}, /* OUT PORT */
	/* 0xe7 */ {dontexist}, /* OUT PORT */

	/* 0xe8 */ {call_dir_handler}, /* CALL DIRECT, 5 bytes */

	/* 0xe9 */ {jmp_dir_handler, INSTR_LEN(5) /*jmp32b no_mem, 5 */ },	/* JMP 32b, 5 bytes */

	/* 0xea */ {dontexist},

	/* 0xeb */ {jmp_dir_handler, INSTR_LEN(2) /*jmp8b no_mem, 2 */ },	/* JMP 8b,2 bytes */

	/* 0xec */ {dontexist},
	/* 0xed */ {dontexist},

	/* 0xee */ {privileged}, /* OUT PORT */
	/* 0xef */ {privileged}, /* OUT PORT */

	/* 0xf0 */ {Plock}, /* LOCK prefix */

	/* 0xf1 */ {dontexist},

	/* 0xf2 */ {rep, 0}, /* REPNE */
	/* 0xf3 */ {rep, 1}, /* REPE/REP, SSE[TODO] */

	/* 0xf4 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* HLT */

	/* 0xf5 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CMC */

	/* 0xf6 */ {check_mod_op2b, IMM_LEN(1), MEM_LEN(1), TYPE(LOAD)}, /* DIV,MUL,NEG,NOT,TEST */
	/* 0xf7 */ {check_mod_op2b, IMM_LEN(4), MEM_LEN(4), TYPE(LOAD)}, /* DIV,MUL,NEG,NOT,TEST */

	/* 0xf8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLC */
	/* 0xf9 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STC */
	/* 0xfa */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLI */
	/* 0xfb */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STI */
	/* 0xfc */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLD */
	/* 0xfd */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STD */

	/* 0xfe */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* inc mem8, dec mem8 */
	/* 0xff */ {opcodeFF} /* inc mem32, dec mem32, PUSH MEM, call *mem, jmp *mem */
};

void X_fetch(unsigned addr){
	address = addr;
	prefix_c = 0;
	immsize_override = 0;
	byte1 = GET_CHAR(address);
	byte2 = GET_CHAR(address + 1);
	byte3 = GET_CHAR(address + 2);
	byte4 = GET_CHAR(address + 3);
}

void skip(){
	prefix_c++;
	byte1 = byte2;
	byte2 = byte3;
	byte3 = byte4;
	byte4 = 0;
}

///////////////////////////////////////////////

unsigned Poperand(){

	switch (byte2) {
		case 0xA0:
			return todo(DUMMY, DUMMY, DUMMY);
		case 0xA1:
			prefix_c++;
			return movl_eax_mem_handler(INSTR_LEN(5), MEM_LEN(2), TYPE(LOAD));
		case 0xA2:
			return todo(DUMMY, DUMMY, DUMMY);
		case 0xA3:
			prefix_c++;
			return movl_eax_mem_handler(INSTR_LEN(5), MEM_LEN(2), TYPE(STORE));
	}

	skip();

	immsize_override = 2;
	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);
}

///////////////////////////////////////////////

unsigned Plock(){

	lock = 1;
	skip();

	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);
}

///////////////////////////////////////////////

unsigned PSegOverride(unsigned reg, unsigned dummy1, unsigned dummy2) {

	seg_override = reg;
	skip();

	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);

}

///////////////////////////////////////////////


decode opcode0F_table[256] = {
	/* 0x00 */ {privileged},				/* LIDT, LLDT, LTR, STR, VERR, VERW, SLDT */
	/* 0x01 */ {privileged},				/* INVLPG, LGDT, LMSW, SMSW */

	/* 0x02 */ {privileged},				/* LAR */
	/* 0x03 */ {privileged},				/* LSL */

	/* 0x04 */ {dontexist},
	/* 0x05 */ {dontexist},

	/* 0x06 */ {privileged},				/* CLTS */

	/* 0x07 */ {dontexist},

	/* 0x08 */ {privileged},				/* INVD */
	/* 0x09 */ {privileged},				/* WBINV */

	/* 0x0a */ {dontexist},

	/* 0x0b */ {dontexist},				/* UB2 */

	/* 0x0c */ {dontexist},

	/* 0x0d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* 3DNow: PREFETCH */
	/* 0x0e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* 3DNow: FEMMS */
	/* 0x0f */ {_3dnow},				/* 3DNow */

	/* 0x10 */ {_sse},				/* MOVUPS */
	/* 0x11 */ {_sse},				/* MOVUPS */
	/* 0x12 */ {_sse},				/* MOVHLPS, MOVLPS */
	/* 0x13 */ {_sse},				/* MOVLPS */
	/* 0x14 */ {_sse},				/* UNPCKLPS */
	/* 0x15 */ {_sse},				/* UNPCKHPS */
	/* 0x16 */ {_sse},				/* MOVHPS, MOVLHPS */
	/* 0x17 */ {_sse},				/* MOVHPS */
	/* 0x18 */ {_sse},				/* PREFETCH. */

	/* 0x19 */ {dontexist},
	/* 0x1a */ {dontexist},
	/* 0x1b */ {dontexist},
	/* 0x1c */ {dontexist},
	/* 0x1d */ {dontexist},
	/* 0x1e */ {dontexist},
	/* 0x1f */ {dontexist},

	/* 0x20 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x21 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x22 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x23 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */

	/* 0x24 */ {dontexist},
	/* 0x25 */ {dontexist},
	/* 0x26 */ {dontexist},
	/* 0x27 */ {dontexist},

	/* 0x28 */ {_sse},				/* MOVAPS */
	/* 0x29 */ {_sse},				/* MOVAPS */
	/* 0x2a */ {_sse},				/* CVTPI2PS */
	/* 0x2b */ {_sse},				/* MOVNTPS */
	/* 0x2c */ {_sse},				/* CVTTPS2PI */
	/* 0x2d */ {_sse},				/* CVTPS2PI */
	/* 0x2e */ {_sse},				/* UCOMISS */
	/* 0x2f */ {_sse},				/* COMISS */

	/* 0x30 */ {privileged},				/* WRMSR */

	/* 0x31 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDTSC */
	/* 0x32 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDMSR */
	/* 0x33 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDPMC */

	/* 0x34 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* SYSENTER */
	/* 0x35 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* SYSEXIT */

	/* 0x36 */ {dontexist},
	/* 0x37 */ {dontexist},
	/* 0x38 */ {dontexist},
	/* 0x39 */ {dontexist},
	/* 0x3a */ {dontexist},
	/* 0x3b */ {dontexist},
	/* 0x3c */ {dontexist},
	/* 0x3d */ {dontexist},
	/* 0x3e */ {dontexist},
	/* 0x3f */ {dontexist},

	/* 0x40 */ {cmovcc},				/* CMOVcc */
	/* 0x41 */ {cmovcc},				/* CMOVcc */
	/* 0x42 */ {cmovcc},				/* CMOVcc */
	/* 0x43 */ {cmovcc},				/* CMOVcc */
	/* 0x44 */ {cmovcc},				/* CMOVcc */
	/* 0x45 */ {cmovcc},				/* CMOVcc */
	/* 0x46 */ {cmovcc},				/* CMOVcc */
	/* 0x47 */ {cmovcc},				/* CMOVcc */
	/* 0x48 */ {cmovcc},				/* CMOVcc */
	/* 0x49 */ {cmovcc},				/* CMOVcc */
	/* 0x4a */ {cmovcc},				/* CMOVcc */
	/* 0x4b */ {cmovcc},				/* CMOVcc */
	/* 0x4c */ {cmovcc},				/* CMOVcc */
	/* 0x4d */ {cmovcc},				/* CMOVcc */
	/* 0x4e */ {cmovcc},				/* CMOVcc */
	/* 0x4f */ {cmovcc},				/* CMOVcc */

	/* 0x50 */ {_sse},				/* MOVMSKPS */
	/* 0x51 */ {_sse},				/* SQRTPS */
	/* 0x52 */ {_sse},				/* RSQRTPS */
	/* 0x53 */ {_sse},				/* RCPPS */
	/* 0x54 */ {_sse},				/* ANDPS */
	/* 0x55 */ {_sse},				/* ANDNPS */
	/* 0x56 */ {_sse},				/* ORPS */
	/* 0x57 */ {_sse},				/* XORPS */
	/* 0x58 */ {_sse},				/* ADDPS */
	/* 0x59 */ {_sse},				/* MULPS */

	/* 0x5a */ {dontexist},
	/* 0x5b */ {dontexist},

	/* 0x5c */ {_sse},				/* SUBPS */
	/* 0x5d */ {_sse},				/* MINPS */

	/* 0x5e */ {dontexist},

	/* 0x5f */ {_sse},				/* MAXPS */

	/* 0x60 */ {_mmx},
	/* 0x61 */ {_mmx},				/* SSE: SQRTPS!!! */
	/* 0x62 */ {_mmx},
	/* 0x63 */ {_mmx},
	/* 0x64 */ {_mmx},
	/* 0x65 */ {_mmx},
	/* 0x66 */ {_mmx},
	/* 0x67 */ {_mmx},
	/* 0x68 */ {_mmx},
	/* 0x69 */ {_mmx},
	/* 0x6a */ {_mmx},
	/* 0x6b */ {_mmx},

	/* 0x6c */ {dontexist},
	/* 0x6d */ {dontexist},

	/* 0x6e */ {_mmx},
	/* 0x6f */ {_mmx},

	/* 0x70 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 *//* SSE: PSHUFW!! */
	/* 0x71 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */
	/* 0x72 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */
	/* 0x73 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */

	/* 0x74 */ {_mmx},
	/* 0x75 */ {_mmx},
	/* 0x76 */ {_mmx},

	/* 0x77 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* MMX: EMMS */

	/* 0x78 */ {dontexist},
	/* 0x79 */ {dontexist},
	/* 0x7a */ {dontexist},
	/* 0x7b */ {dontexist},
	/* 0x7c */ {dontexist},
	/* 0x7d */ {dontexist},

	/* 0x7e */ {_mmx},
	/* 0x7f */ {_mmx},

	/* 0x80 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x81 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x82 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x83 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x84 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x85 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x86 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x87 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x88 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x89 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8a */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8b */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8c */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8d */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8e */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8f */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */

	/* 0x90 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x91 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x92 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x93 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x94 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x95 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x96 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x97 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x98 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x99 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9b */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9c */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9d */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9e */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9f */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */

	/* 0xa0 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* PUSH FS */
	/* 0xa1 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* POP FS */

	/* 0xa2 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* CPUID */

	/* 0xa3 */ {btx_handler, TYPE(LOAD)},                 /* BT reg, r/m */

	/* 0xa4 */ {todo},				/* SHLD IMM8 */
	/* 0xa5 */ {todo},				/* SHLD (CL) */

	/* 0xa6 */ {dontexist},
	/* 0xa7 */ {dontexist},

	/* 0xa8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* PUSH GS */
	/* 0xa9 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* POP GS */

	/* 0xaa */ {privileged},				/* RSM */

	/* 0xab */ {btx_handler,TYPE(MODIFY)},                /* BTS reg, r/m */

	/* 0xac */ {no_mem, INSTR_LEN(3), IMM_LEN(1) /*todo */ },				/* SHRD IMM8 *///GOGO
	/* 0xad */ {no_mem, INSTR_LEN(2), IMM_LEN(0) /*todo */ },				/* SHRD (CL) *///GOGO

	/* 0xae */ {_sse},				/* FXRSTOR, FXSAVE, LDMXCSR, STMXCSR, SFENCE */

	/* 0xaf */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)},				/* IMUL reg,mem */

	/* 0xb0 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)},				/* CMPXCHG */
	/* 0xb1 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)},				/* CMPXCHG */

	/* 0xb2 */ {todo},				/* BTC */

	/* 0xb3 */ {btx_handler, TYPE(MODIFY)},               /* BTR reg, r/m*/

	/* 0xb4 */ {todo},				/* LFS */
	/* 0xb5 */ {todo},				/* LGS */

	/* 0xb6 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},	/* MOVZBL, MOVZBW met Poperand */
	/* 0xb7 */ {check_mod, IMM_LEN(0), MEM_LEN(2), TYPE(LOAD)},	/* MOVZWL */

	/* 0xb8 */ {dontexist},
	/* 0xb9 */ {dontexist},

	/* 0xba */ {todo},				/* BT, BTC, BTS */
	/* 0xbb */ {todo},				/* BTC */
	/* 0xbc */ {todo},				/* BSF */
	/* 0xbd */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* BSR */// GOGO

	/* 0xbe */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)},	/* MOVSX, nooit een immediate! */
	/* 0xbf */ {check_mod, IMM_LEN(0), MEM_LEN(2), TYPE(LOAD)},	/* MOVSX */

	/* 0xc0 */ {todo},						/* XADD */
	/* 0xc1 */ {todo},						/* XADD */

	/* 0xc2 */ {_sse},						/* CMPPS */

	/* 0xc3 */ {dontexist},

	/* 0xc4 */ {_sse},				/* PEXTRW */
	/* 0xc5 */ {_sse},				/* PEXTRW */
	/* 0xc6 */ {_sse},				/* SHUFPS */

	/* 0xc7 */ {todo},				/* CMPXCHG8B */

	/* 0xc8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				// bswap %eax
	/* 0xc9 */ {dontexist},
	/* 0xca */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				// bswap %edx
	/* 0xcb */ {dontexist},
	/* 0xcc */ {dontexist},
	/* 0xcd */ {dontexist},
	/* 0xce */ {dontexist},
	/* 0xcf */ {dontexist},

	/* 0xd0 */ {_mmx},
	/* 0xd1 */ {_mmx},
	/* 0xd2 */ {_mmx},
	/* 0xd3 */ {_mmx},

	/* 0xd4 */ {dontexist},

	/* 0xd5 */ {_mmx},

	/* 0xd6 */ {dontexist},

	/* 0xd7 */ {_sse},				/* PMOVMSKB */

	/* 0xd8 */ {_mmx},
	/* 0xd9 */ {_mmx},
	/* 0xda */ {_mmx},				/* SSE: PMINUB!! */
	/* 0xdb */ {_mmx},
	/* 0xdc */ {_mmx},
	/* 0xdd */ {_mmx},
	/* 0xde */ {_mmx},				/* SSE: PMAXUB!! */
	/* 0xdf */ {_mmx},
	/* 0xe0 */ {_mmx},				/* SSE: PAVGB!! */
	/* 0xe1 */ {_mmx},
	/* 0xe2 */ {_mmx},
	/* 0xe3 */ {_mmx},				/* SSE: PAVGB!! */

	/* 0xe4 */ {_sse},				/* PMULHUW */

	/* 0xe5 */ {_mmx},

	/* 0xe6 */ {dontexist},

	/* 0xe7 */ {_sse},				/* MOVNTQ */

	/* 0xe8 */ {_mmx},
	/* 0xe9 */ {_mmx},
	/* 0xea */ {_mmx},				/* SSE: PWINSW!! */
	/* 0xeb */ {_mmx},
	/* 0xec */ {_mmx},
	/* 0xed */ {_mmx},
	/* 0xee */ {_mmx},				/* SSE: PMAXSW!! */
	/* 0xef */ {_mmx},
	/* 0xf0 */ {_mmx},
	/* 0xf1 */ {_mmx},
	/* 0xf2 */ {_mmx},
	/* 0xf3 */ {_mmx},

	/* 0xf4 */ {dontexist},

	/* 0xf5 */ {_mmx},

	/* 0xf6 */ {_sse},				/* PSADBW */
	/* 0xf7 */ {_sse},				/* MASKMOVQ */

	/* 0xf8 */ {_mmx},
	/* 0xf9 */ {_mmx},
	/* 0xfa */ {_mmx},
	/* 0xfb */ {_mmx},
	/* 0xfc */ {_mmx},
	/* 0xfd */ {_mmx},
	/* 0xfe */ {_mmx},
	/* 0xff */ {_mmx}
};

unsigned extended0F(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	skip();
	return opcode0F_table[byte1].function(opcode0F_table[byte1].arg1,
			opcode0F_table[byte1].arg2,
			opcode0F_table[byte1].arg3);
}

/**********************************************************************************************/

unsigned LOOKUP_CLONE2INSTR(unsigned clone_address){
	unsigned i=0;
	unsigned result;

	LOCK_CLONE2INSTR();
	for (i=0; i< __instr_p__; i++)
		if (_clone2instr_[i].from==clone_address) {
			result = _clone2instr_[i].to;
			UNLOCK_CLONE2INSTR();
			return result;
		}
	UNLOCK_CLONE2INSTR();
	
	return 0;
}


/**********************************************************************************************/

unsigned LOOKUP_RELOC(unsigned address){
	int i = 0;
	unsigned result;

	if (!GEBRUIK_RELOC_INFO)
		return 0;

	LOCK_RELOC();
#ifdef USE_HASH
	DBG(fprintf(jiti_tty, "\n>>>[LOOKUP %x ? %x]", address));
	result = (_reloc_info_[HASH(address)]._from_ ==
			address) ? _reloc_info_[HASH(address)]._to_ : 0;
	UNLOCK_RELOC();
	return result;
#else
	DBG(fprintf(jiti_tty, "\n>>>[LOOKUP %x]", address));
	while (i < RELOC) {
		if (_reloc_info_[i]._from_ == address) {
			result = _reloc_info_[i]._to_;
			UNLOCK_RELOC();
			return result;
		}
		i++;
	}
	DBG(fprintf(jiti_tty, "\n>>>NOT FOUND"));
	UNLOCK_RELOC();
	return 0;			/* niet gevonden */
#endif

}

/**********************************************************************************************/

unsigned decode_instruction(char show){
	X_len = -1;
	X_kloon = 0;
	seg_override = 0;
	if (lock)
		panic("unprocessed lock!!!");

	X_fetch(X_address);
	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);

}

/**********************************************************************************************/

#ifdef DEBUG
unsigned main_ret_address;
#endif //DEBUG

void jiti_restart_instr();

unsigned decode_block(unsigned address){
	unsigned X_instr_start = (unsigned) X_instr;
	DBG(unsigned X_address1);
	DBG(unsigned X_instr1);


	LOCK_DECODER();

	// OK, par -> we hebben de lock

	address -= delta;
	X_address = address;

	LOCK_CLONE2INSTR();
	_clone2instr_[__instr_p__].from = X_address + delta;
	_clone2instr_[__instr_p__].to = (int) X_instr;
	DBG(fprintf(jiti_tty, "\n>>>CLONE2INSTR: %x -> %x", 
				_clone2instr_[__instr_p__].from,
				_clone2instr_[__instr_p__].to));
	__instr_p__++;
	UNLOCK_CLONE2INSTR();

	if (_instrument_code_) _X_start=X_address;

	// save the return address of main so that we know when we have to
	// stop instrumenting
	if (X_instr == X_instr_) {
#ifdef DEBUG
		PUT_SHORT(X_instr,0x2589);
		X_instr+=2;
		PUT_LONG(X_instr,&main_ret_address);
		X_instr+=4;  // movl %esp,main_ret_address
#endif //DEBUG
		PUSH_REG(EAX);
		X_instr++;
		PUT_LONG(X_instr,0x0424448b);
		X_instr+=4;  // movl 4(%esp),%eax
		MOVL_EAX_2_MEM(X_instr,&main_ret);
		X_instr+=5;
		POP_REG(EAX);
		X_instr++;
	}

	while (X_address) {

#ifdef DEBUG
		fprintf(jiti_tty, "\n### ");
		jiti_fprint_ins1(jiti_tty, 1, X_address);
		X_address1 = X_address;
		// update delta since when we follow direct jumps/calls, we can move
		// between the application and dynlib code
#endif

		X_instr_start_block = (int) X_instr;
		if (GEBRUIK_RELOC_INFO) {
			UPDATE_RELOC(X_address, X_instr);
		}
		X_address = decode_instruction(1);

#ifdef DEBUG
		fprintf(jiti_tty, "\n*** ");
		jiti_fprint_ins1(jiti_tty, 1, X_address1 + delta);
		if (X_instr1 != X_instr) {
			fprintf(jiti_tty, "\n+++ ");
			jiti_fprint_ins(jiti_tty, X_instr1, X_instr);
		}
		X_instr1 = X_instr;
#endif

	}

	if (GEBRUIK_PATCH_INFO) {
		unsigned i, new_target;
		DBG(fprintf(jiti_tty, "\n>>>TO BE PATCHED: %d ", to_be_patched_i));
		for (i = 0; i < to_be_patched_i; i++) {
			DBG(fprintf(jiti_tty, " PATCH: %x:%x? ",
						_to_be_patched_info_[i]._address_,
						_to_be_patched_info_[i]._target_));
			if ((new_target=LOOKUP_RELOC(_to_be_patched_info_[i]._target_))){

				DBG(fprintf(jiti_tty, " PATCH: %x: %x -> %x",
							_to_be_patched_info_[i]._address_,
							GET_LONG(_to_be_patched_info_[i]._address_) +
							_to_be_patched_info_[i]._address_,
							new_target));

				/*
				 * DBG(jiti_fprint_ins(jiti_tty, _to_be_patched_info_[i]._address_-2, 
				 * _to_be_patched_info_[i]._address_+4));
				 */

				*(unsigned *) _to_be_patched_info_[i]._address_ =
					new_target - _to_be_patched_info_[i]._address_ - 4;

			} else if ((new_target = LOOKUP_CLONE2INSTR(_to_be_patched_info_[i]._target_))) {

				DBG(fprintf(jiti_tty, " found in CLONE2INSTR: %x: %x -> %x",
							_to_be_patched_info_[i]._address_,
							GET_LONG(_to_be_patched_info_[i]._address_) +
							_to_be_patched_info_[i]._address_, new_target));

				//DBG(jiti_fprint_ins(jiti_tty, _to_be_patched_info_[i]._address_-2, _to_be_patched_info_[i]._address_+4));
				*(unsigned *) _to_be_patched_info_[i]._address_ =
					new_target - _to_be_patched_info_[i]._address_ - 4;
			}

		}
	}

	UNLOCK_DECODER();

	return X_instr_start;

}

/********************************************************************************************/

#define INITIAL_PAR_SIZE 128*1024*1024
#define INITIAL_SEQ_SIZE 32*1024*1024	// minimaal 8kB !!
#define REDZONE_FACTOR   4
unsigned x_instr_size = 0;
unsigned x_redzone = 0;

time_t time_of_last_realloc = 0;

/////

void check_x_instr_size() {
	// we lopen in REDZONE => opkuisen en herstarten
	if ((X_instr + x_redzone) > (X_instr_ + x_instr_size))
		jiti_restart_instr();
}

////

int jiti_parallel_program();	// jiti.c

void jiti_start_instr(){

	time(&time_of_last_realloc);

	x_instr_size =
		jiti_parallel_program()? INITIAL_PAR_SIZE : INITIAL_SEQ_SIZE;
	x_redzone = x_instr_size / REDZONE_FACTOR;

	if (!(X_instr_ = (char *) malloc(x_instr_size))) panic("malloc() failed\n");

	DBG(fprintf(jiti_tty, "X_instr_: Allocating %d bytes from 0x%x to 0x%x @ %d, "
				"REDZONE (%d bytes) starts at 0x%x\n", x_instr_size,
				(unsigned) X_instr_, (unsigned) X_instr_ + x_instr_size,
				(unsigned) time_of_last_realloc, x_redzone,
				(unsigned) X_instr_ + x_redzone));
	X_instr = X_instr_;
}

/////

void jiti_restart_instr(){
	time_t new_time;

	time(&new_time);

	DBG(fprintf
			(jiti_tty, "X_instr_: RESTARTING @ %d\n", (unsigned) new_time));

	fprintf(stderr, "X_instr_: RESTARTING @ %d\n", (unsigned) new_time);

	if (jiti_parallel_program())
		panic("Reinstrumenting parallel program!\n");

	// seq. program

	DBG(fprintf
			(jiti_tty, "X_instr_: Freeing    %d bytes @ 0x%x\n", x_instr_size,
			 (unsigned) X_instr_));
	free(X_instr_);

	// nieuwe grootte voor X_instr_ bepalen
	{
#define MAKE_SMALLER 60
#define MAKE_BIGGER  10

		/*
		 * MAKE_BIGGER                        MAKE_SMALLER
		 * --------------+----------------------------------+---------------------->
		 * 
		 * *2                      =                            /2
		 * 
		 */

		if (new_time > (time_of_last_realloc + MAKE_SMALLER)) {
			x_instr_size >>= 1;	// /2
		} else if (new_time < (time_of_last_realloc + MAKE_BIGGER)) {
			x_instr_size <<= 1;	// *2
		}
		x_redzone = x_instr_size / REDZONE_FACTOR;
		time_of_last_realloc = new_time;
	}

	// clone2instr invalideren
	__instr_p__ = 0;

	// _reloc_info_ invalideren
	{
		unsigned i;
		for (i = 0; i < RELOC; i++) {
			_reloc_info_[i]._from_ = 0;
			_reloc_info_[i]._to_ = 0;
		}
	}

	// _to_be_patched_info_ invalideren
	{
		unsigned i;
		for (i = 0; i < PATCH; i++) {
			_to_be_patched_info_[i]._address_ = 0;
			_to_be_patched_info_[i]._target_ = 0;
		}
	}

	// nieuwe X_instr_ aanmaken
	if (!(X_instr_ = (char *) malloc(x_instr_size)))
		panic("malloc() failed\n");
	DBG(fprintf
			(jiti_tty,
			 "X_instr_: Allocating %d bytes @ 0x%x REDZONE (%d bytes) starts at 0x%x\n",
			 x_instr_size, (unsigned) X_instr_, x_redzone,
			 (unsigned) X_instr_ + x_redzone));
	X_instr = X_instr_;
}
